#!/usr/bin/python
from collections import defaultdict

from Qt import QtWidgets, QtCore, QtGui

from ..constants import (NODE_PROP_QLABEL,
                                   NODE_PROP_QLINEEDIT,
                                   NODE_PROP_QTEXTEDIT,
                                   NODE_PROP_QCOMBO,
                                   NODE_PROP_QCHECKBOX,
                                   NODE_PROP_QSPINBOX,
                                   NODE_PROP_COLORPICKER,
                                   NODE_PROP_SLIDER,
                                   NODE_PROP_FILE,
                                   NODE_PROP_FILE_SAVE,
                                   NODE_PROP_VECTOR2,
                                   NODE_PROP_VECTOR3,
                                   NODE_PROP_VECTOR4,
                                   NODE_PROP_FLOAT,
                                   NODE_PROP_INT,
                                   NODE_PROP_BUTTON)
from ..widgets.dialogs import FileDialog


class BaseProperty(QtWidgets.QWidget):
    """
    Base widget class for a node property.
    """

    value_changed = QtCore.Signal(str, object)

    def set_value(self, value):
        raise NotImplementedError

    def get_value(self):
        raise NotImplementedError


class PropColorPicker(BaseProperty):
    """
    Color picker widget for a node property.
    """

    def __init__(self, parent=None):
        super(PropColorPicker, self).__init__(parent)
        self._color = (0, 0, 0)
        self._button = QtWidgets.QPushButton()
        # self._vector = PropVector3()
        # self._vector.set_value([0, 0, 0])
        self._update_color()

        self._button.clicked.connect(self._on_select_color)
        # self._vector.value_changed.connect(self._on_vector_changed)
        layout = QtWidgets.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._button, 0, QtCore.Qt.AlignLeft)
        # layout.addWidget(self._vector, 1, QtCore.Qt.AlignLeft)

    # def _on_vector_changed(self, o, value):
    #     self._color = tuple(value)
    #     self._update_color()
    #     self.value_changed.emit(self.toolTip(), value)

    def _on_select_color(self):
        options = QtWidgets.QColorDialog.ColorDialogOptions()
        # options |= QtWidgets.QColorDialog.ShowAlphaChannel
        options |= QtWidgets.QColorDialog.DontUseNativeDialog
        dlg = QtWidgets.QColorDialog(QtGui.QColor.fromRgb(*self.get_value()))
        color = dlg.getColor(
            QtGui.QColor.fromRgb(*self.get_value()), options=options)

        if color.isValid():
            self.set_value(color.getRgb())

    # def _update_vector(self):
    #     self._vector.set_value(list(self._color))

    def _update_color(self):
        c = [int(max(min(i, 255), 0)) for i in self._color]
        hex_color = '#{0:02x}{1:02x}{2:02x}'.format(*c)
        self._button.setStyleSheet(
            '''
            QPushButton {{background-color: rgba({0}, {1}, {2}, 255);}}
            QPushButton::hover {{background-color: rgba({0}, {1}, {2}, 200);}}
            '''.format(*c)
        )
        self._button.setToolTip(
            'rgb: {}\nhex: {}'.format(self._color[:3], hex_color)
        )

    def get_value(self):
        return self._color[:3]

    def set_value(self, value):
        if value != self.get_value():
            self._color = value
            self._update_color()
            # self._update_vector()
            self.value_changed.emit(self.toolTip(), value)


class PropSlider(BaseProperty):
    """
    Slider widget for a node property.
    """

    def __init__(self, parent=None):
        super(PropSlider, self).__init__(parent)
        self._block = False
        self._slider = QtWidgets.QSlider()
        self._spnbox = QtWidgets.QSpinBox()
        self._init()

    def _init(self):
        self._slider.setOrientation(QtCore.Qt.Horizontal)
        self._slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self._slider.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                   QtWidgets.QSizePolicy.Preferred)
        self._spnbox.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)
        layout = QtWidgets.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._spnbox)
        layout.addWidget(self._slider)
        self._spnbox.valueChanged.connect(self._on_spnbox_changed)
        self._slider.valueChanged.connect(self._on_slider_changed)
        # store the original press event.
        self._slider_mouse_press_event = self._slider.mousePressEvent
        self._slider.mousePressEvent = self._on_slider_mouse_press
        self._slider.mouseReleaseEvent = self._on_slider_mouse_release

    def _on_slider_mouse_press(self, event):
        self._block = True
        self._slider_mouse_press_event(event)

    def _on_slider_mouse_release(self, event):
        self.value_changed.emit(self.toolTip(), self.get_value())
        self._block = False

    def _on_slider_changed(self, value):
        self._spnbox.setValue(value)

    def _on_spnbox_changed(self, value):
        if value != self._slider.value():
            self._slider.setValue(value)
            if not self._block:
                self.value_changed.emit(self.toolTip(), self.get_value())

    def get_value(self):
        return self._spnbox.value()

    def set_value(self, value):
        if value != self.get_value():
            self._block = True
            self._spnbox.setValue(value)
            self.value_changed.emit(self.toolTip(), value)
            self._block = False

    def set_min(self, value=0):
        self._spnbox.setMinimum(value)
        self._slider.setMinimum(value)

    def set_max(self, value=0):
        self._spnbox.setMaximum(value)
        self._slider.setMaximum(value)


class PropLabel(QtWidgets.QLabel):
    """
    Label widget for a node property.
    """

    value_changed = QtCore.Signal(str, object)

    def get_value(self):
        return self.text()

    def set_value(self, value):
        if value != self.get_value():
            self.setText(str(value))
            self.value_changed.emit(self.toolTip(), value)


class PropLineEdit(QtWidgets.QLineEdit):
    """
    LineEdit widget for a node property.
    """

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(PropLineEdit, self).__init__(parent)
        self.editingFinished.connect(self._on_editing_finished)

    def _on_editing_finished(self):
        self.value_changed.emit(self.toolTip(), self.text())

    def get_value(self):
        return self.text()

    def set_value(self, value):
        _value = str(value)
        if _value != self.get_value():
            self.setText(_value)
            self.value_changed.emit(self.toolTip(), _value)


class PropTextEdit(QtWidgets.QTextEdit):
    """
    TextEdit widget for a node property.
    """

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(PropTextEdit, self).__init__(parent)
        self.__prev_text = ''

    def focusInEvent(self, event):
        super(PropTextEdit, self).focusInEvent(event)
        self.__prev_text = self.toPlainText()

    def focusOutEvent(self, event):
        super(PropTextEdit, self).focusOutEvent(event)
        if self.__prev_text != self.toPlainText():
            self.value_changed.emit(self.toolTip(), self.toPlainText())
        self.__prev_text = ''

    def get_value(self):
        return self.toPlainText()

    def set_value(self, value):
        _value = str(value)
        if _value != self.get_value():
            self.setPlainText(_value)
            self.value_changed.emit(self.toolTip(), _value)


class PropComboBox(QtWidgets.QComboBox):
    """
    ComboBox widget for a node property.
    """

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(PropComboBox, self).__init__(parent)
        self.currentIndexChanged.connect(self._on_index_changed)

    def _on_index_changed(self):
        self.value_changed.emit(self.toolTip(), self.get_value())

    def items(self):
        """
        returns items from the combobox.

        Returns:
            list[str]: list of strings.
        """
        return [self.itemText(i) for i in range(self.count())]

    def set_items(self, items):
        """
        Set items on the combobox.

        Args:
            items (list[str]): list of strings.
        """
        self.clear()
        self.addItems(items)

    def get_value(self):
        return self.currentText()

    def set_value(self, value):
        if value != self.get_value():
            idx = self.findText(value, QtCore.Qt.MatchExactly)
            self.setCurrentIndex(idx)
            if idx >= 0:
                self.value_changed.emit(self.toolTip(), value)


class PropCheckBox(QtWidgets.QCheckBox):
    """
    CheckBox widget for a node property.
    """

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(PropCheckBox, self).__init__(parent)
        self.clicked.connect(self._on_clicked)

    def _on_clicked(self):
        self.value_changed.emit(self.toolTip(), self.get_value())

    def get_value(self):
        return self.isChecked()

    def set_value(self, value):
        if value != self.get_value():
            self.setChecked(value)
            self.value_changed.emit(self.toolTip(), value)


class PropSpinBox(QtWidgets.QSpinBox):
    """
    SpinBox widget for a node property.
    """

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(PropSpinBox, self).__init__(parent)
        self.setButtonSymbols(self.NoButtons)
        self.valueChanged.connect(self._on_value_change)

    def _on_value_change(self, value):
        self.value_changed.emit(self.toolTip(), value)

    def get_value(self):
        return self.value()

    def set_value(self, value):
        if value != self.get_value():
            self.setValue(value)


class PropFilePath(BaseProperty):

    def __init__(self, parent=None):
        super(PropFilePath, self).__init__(parent)
        self._ledit = QtWidgets.QLineEdit()
        self._ledit.setAlignment(QtCore.Qt.AlignLeft)
        self._ledit.editingFinished.connect(self._on_value_change)
        self._ledit.clearFocus()

        icon = self.style().standardIcon(QtWidgets.QStyle.StandardPixmap(21))
        _button = QtWidgets.QPushButton()
        _button.setIcon(icon)
        _button.clicked.connect(self._on_select_file)

        hbox = QtWidgets.QHBoxLayout(self)
        hbox.setContentsMargins(0, 0, 0, 0)
        hbox.addWidget(self._ledit)
        hbox.addWidget(_button)

        self._ext = '*'
        self._file_directory = None

    def _on_select_file(self):
        file_path = FileDialog.getOpenFileName(self,
                                               file_dir=self._file_directory,
                                               ext_filter=self._ext)
        file = file_path[0] or None
        if file:
            self.set_value(file)

    def _on_value_change(self, value=None):
        if value is None:
            value = self._ledit.text()
        self.value_changed.emit(self.toolTip(), value)

    def set_file_ext(self, ext=None):
        self._ext = ext or '*'

    def set_file_directory(self, directory):
        self._file_directory = directory

    def get_value(self):
        return self._ledit.text()

    def set_value(self, value):
        _value = str(value)
        if _value != self.get_value():
            self._ledit.setText(_value)
            self._on_value_change(_value)


class PropFileSavePath(PropFilePath):

    def _on_select_file(self):
        file_path = FileDialog.getSaveFileName(self,
                                               file_dir=self._file_directory,
                                               ext_filter=self._ext)
        file = file_path[0] or None
        if file:
            self.set_value(file)


class _ValueMenu(QtWidgets.QMenu):

    mouseMove = QtCore.Signal(object)
    mouseRelease = QtCore.Signal(object)
    stepChange = QtCore.Signal()

    def __init__(self, parent=None):
        super(_ValueMenu, self).__init__(parent)
        self.step = 1
        self.last_action = None
        self.steps = []

    def set_steps(self, steps):
        self.clear()
        self.steps = steps
        for step in steps:
            self._add_action(step)

    def _add_action(self, step):
        action = QtWidgets.QAction(str(step), self)
        action.step = step
        self.addAction(action)

    def mouseMoveEvent(self, event):
        self.mouseMove.emit(event)
        super(_ValueMenu, self).mouseMoveEvent(event)

        action = self.actionAt(event.pos())
        if action:
            if action is not self.last_action:
                self.stepChange.emit()
            self.last_action = action
            self.step = action.step
        elif self.last_action:
            self.setActiveAction(self.last_action)

    def mousePressEvent(self, event):
        return

    def mouseReleaseEvent(self, event):
        self.mouseRelease.emit(event)
        super(_ValueMenu, self).mouseReleaseEvent(event)

    def set_data_type(self, dt):
        if dt is int:
            new_steps = []
            for step in self.steps:
                if '.' not in str(step):
                    new_steps.append(step)
            self.set_steps(new_steps)
        elif dt is float:
            self.set_steps(self.steps)


class _ValueEdit(QtWidgets.QLineEdit):

    valueChanged = QtCore.Signal(object)

    def __init__(self, parent=None):
        super(_ValueEdit, self).__init__(parent)
        self.mid_state = False
        self._data_type = float
        self.setText('0')

        self.pre_x = None
        self.pre_val = None
        self._step = 1
        self._speed = 0.1

        self.editingFinished.connect(self._on_text_changed)

        self.menu = _ValueMenu()
        self.menu.mouseMove.connect(self.mouseMoveEvent)
        self.menu.mouseRelease.connect(self.mouseReleaseEvent)
        self.menu.stepChange.connect(self._reset)
        steps = [0.001, 0.01, 0.1, 1, 10, 100, 1000]
        self.menu.set_steps(steps)

        self.set_data_type(float)

    def _on_text_changed(self):
        self.valueChanged.emit(self.value())

    def _reset(self):
        self.pre_x = None

    def mouseMoveEvent(self, event):
        if self.mid_state:
            if self.pre_x is None:
                self.pre_x = event.x()
                self.pre_val = self.value()
            else:
                self.set_step(self.menu.step)
                delta = event.x() - self.pre_x
                value = self.pre_val + int(delta * self._speed) * self._step
                self.setValue(value)
                self._on_text_changed()

        super(_ValueEdit, self).mouseMoveEvent(event)

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.MiddleButton:
            self.mid_state = True
            self._reset()
            self.menu.exec_(QtGui.QCursor.pos())
        super(_ValueEdit, self).mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        self.menu.close()
        self.mid_state = False
        super(_ValueEdit, self).mouseReleaseEvent(event)

    def set_step(self, step):
        self._step = step

    def set_data_type(self, dt):
        if dt is int:
            self.setValidator(QtGui.QIntValidator())
        elif dt is float:
            self.setValidator(QtGui.QDoubleValidator())
        self._data_type = dt
        self.menu.set_data_type(dt)

    def _convert_text(self, text):
        # int("1.0") will return error
        # so we use int(float("1.0"))
        try:
            value = float(text)
        except:
            value = 0.0
        if self._data_type is int:
            value = int(value)
        return value

    def value(self):
        if self.text().startswith('.'):
            text = '0' + self.text()
            self.setText(text)
        return self._convert_text(self.text())

    def setValue(self, value):
        if value != self.value():
            self.setText(str(self._convert_text(value)))


class _Slider(QtWidgets.QSlider):

    def __init__(self, parent=None):
        super(_Slider, self).__init__(parent)
        self.setOrientation(QtCore.Qt.Horizontal)
        self.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                           QtWidgets.QSizePolicy.Preferred)

    def _update_value(self, x):
        value = (self.maximum() - self.minimum()) * x / self.width() + self.minimum()
        self.setValue(value)

    def mousePressEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            self._update_value(event.pos().x())
        super(_Slider, self).mousePressEvent(event)


class _ValueSliderEdit(QtWidgets.QWidget):

    valueChanged = QtCore.Signal(object)

    def __init__(self, parent=None):
        super(_ValueSliderEdit, self).__init__(parent)
        self._edit = _ValueEdit()
        self._edit.valueChanged.connect(self._on_edit_changed)
        self._edit.setMaximumWidth(70)
        self._slider = _Slider()
        self._slider.valueChanged.connect(self._on_slider_changed)

        hbox = QtWidgets.QHBoxLayout()
        hbox.setContentsMargins(0, 0, 0, 0)
        hbox.addWidget(self._edit)
        hbox.addWidget(self._slider)
        self.setLayout(hbox)

        self._mul = 1000.0
        self.set_min(0)
        self.set_max(10)
        self.set_data_type(float)
        self._lock = False

    def _on_edit_changed(self, value):
        self._set_slider_value(value)
        self.valueChanged.emit(self._edit.value())

    def _on_slider_changed(self, value):
        if self._lock:
            self._lock = False
            return
        value = value / float(self._mul)
        self._edit.setValue(value)
        self._on_edit_changed(value)

    def _set_slider_value(self, value):
        value = int(value * self._mul)

        if value == self._slider.value():
            return
        self._lock = True
        _min = self._slider.minimum()
        _max = self._slider.maximum()
        if _min <= value <= _max:
            self._slider.setValue(value)
        elif value < _min and self._slider.value() != _min:
            self._slider.setValue(_min)
        elif value > _max and self._slider.value() != _max:
            self._slider.setValue(_max)

    def set_min(self, value=0):
        self._slider.setMinimum(int(value * self._mul))

    def set_max(self, value=10):
        self._slider.setMaximum(int(value * self._mul))

    def set_data_type(self, dt):
        _min = int(self._slider.minimum() / self._mul)
        _max = int(self._slider.maximum() / self._mul)
        if dt is int:
            self._mul = 1.0
        elif dt is float:
            self._mul = 1000.0

        self.set_min(_min)
        self.set_max(_max)
        self._edit.set_data_type(dt)

    def value(self):
        return self._edit.value()

    def setValue(self, value):
        self._edit.setValue(value)
        self._on_edit_changed(value)


class _DoubleSpinBox(QtWidgets.QDoubleSpinBox):

    def __init__(self, parent=None):
        super(_DoubleSpinBox, self).__init__(parent)
        self.setButtonSymbols(self.NoButtons)
        self.setRange(-9999999999999999.0, 9999999999999999.0)
        self.setDecimals(16)
        self.setValue(0)
        self.setStyleSheet('QDoubleSpinBox{ border:1px solid }')

    def textFromValue(self, value):
        return str(value)


class PropVector(BaseProperty):

    def __init__(self, parent=None, dim=3):
        super(PropVector, self).__init__(parent)
        self._value = []
        self._items = []
        self._can_emit = True

        hbox = QtWidgets.QHBoxLayout(self)
        hbox.setSpacing(2)
        hbox.setContentsMargins(0, 0, 0, 0)
        for i in range(dim):
            self._add_item(i, hbox)

    def set_data_type(self, dt):
        [item.set_data_type(dt) for item in self._items]

    def _add_item(self, index, hbox):
        _ledit = _ValueEdit()
        _ledit.index = index
        _ledit.valueChanged.connect(
            lambda: self._on_value_change(_ledit.value(), _ledit.index)
        )

        hbox.addWidget(_ledit)
        self._value.append(0.0)
        self._items.append(_ledit)

    def _on_value_change(self, value=None, index=None):
        if self._can_emit:
            if index is not None:
                self._value[index] = value
            self.value_changed.emit(self.toolTip(), self._value)
        self.value_changed.emit(self.toolTip(), self._value)

    def _update_items(self):
        for index, value in enumerate(self._value):
            if index < len(self._items) and self._items[index].value() != value:
                self._items[index].setValue(value)

    def get_value(self):
        return self._value

    def set_value(self, value):
        if value != self.get_value():
            self._value = value.copy()
            self._can_emit = False
            self._update_items()
            self._can_emit = True
            self._on_value_change()


class PropVector2(PropVector):

    def __init__(self, parent=None):
        super(PropVector2, self).__init__(parent, 2)


class PropVector3(PropVector):

    def __init__(self, parent=None):
        super(PropVector3, self).__init__(parent, 3)


class PropVector4(PropVector):

    def __init__(self, parent=None):
        super(PropVector4, self).__init__(parent, 4)


class PropFloat(_ValueSliderEdit):

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(PropFloat, self).__init__(parent)
        self.valueChanged.connect(self._on_value_changed)

    def _on_value_changed(self, value):
        self.value_changed.emit(self.toolTip(), value)

    def get_value(self):
        return self.value()

    def set_value(self, value):
        if value != self.get_value():
            self.setValue(value)
            self.value_changed.emit(self.toolTip(), value)


class PropInt(PropFloat):

    def __init__(self, parent=None):
        super(PropInt, self).__init__(parent)
        self.set_data_type(int)


class PropButton(QtWidgets.QPushButton):

    value_changed = QtCore.Signal(str, object)

    def __init__(self, parent=None):
        super(PropButton, self).__init__(parent)

    def set_value(self, value, node=None):
        # value: list of functions
        if type(value) is not list:
            return
        for func in value:
            self.clicked.connect(lambda: func(node))

    def get_value(self):
        return None


WIDGET_MAP = {
    NODE_PROP_QLABEL: PropLabel,
    NODE_PROP_QLINEEDIT: PropLineEdit,
    NODE_PROP_QTEXTEDIT: PropTextEdit,
    NODE_PROP_QCOMBO: PropComboBox,
    NODE_PROP_QCHECKBOX: PropCheckBox,
    NODE_PROP_QSPINBOX: PropSpinBox,
    NODE_PROP_COLORPICKER: PropColorPicker,
    NODE_PROP_SLIDER: PropSlider,

    NODE_PROP_FILE: PropFilePath,
    NODE_PROP_FILE_SAVE: PropFileSavePath,
    NODE_PROP_VECTOR2: PropVector2,
    NODE_PROP_VECTOR3: PropVector3,
    NODE_PROP_VECTOR4: PropVector4,
    NODE_PROP_FLOAT: PropFloat,
    NODE_PROP_INT: PropInt,
    NODE_PROP_BUTTON: PropButton
}


# main property widgets.
# ==============================================================================


class PropListWidget(QtWidgets.QWidget):
    """
    Node properties list displayed under a tab in the NodePropWidget widget.
    """

    def __init__(self, parent=None):
        super(PropListWidget, self).__init__(parent)
        self.__layout = QtWidgets.QGridLayout()
        self.__layout.setColumnStretch(1, 1)
        self.__layout.setSpacing(6)

        layout = QtWidgets.QVBoxLayout(self)
        layout.setAlignment(QtCore.Qt.AlignTop)
        layout.addLayout(self.__layout)

    def __repr__(self):
        return '<{} object at {}>'.format(
            self.__class__.__name__, hex(id(self))
        )

    def add_widget(self, name, widget, value, label=None):
        """
        Add a property widget to the window.

        Args:
            name (str): property name to be displayed.
            widget (BaseProperty): property widget.
            value (object): property value.
            label (str): custom label to display.
        """
        widget.setToolTip(name)
        widget.set_value(value)
        if label is None:
            label = name
        row = self.__layout.rowCount()
        if row > 0:
            row += 1

        label_flags = QtCore.Qt.AlignCenter | QtCore.Qt.AlignRight
        if widget.__class__.__name__ == 'PropTextEdit':
            label_flags = label_flags | QtCore.Qt.AlignTop

        self.__layout.addWidget(QtWidgets.QLabel(label), row, 0, label_flags)
        self.__layout.addWidget(widget, row, 1)

    def get_widget(self, name):
        """
        Returns the property widget from the name.

        Args:
            name (str): property name.

        Returns:
            QtWidgets.QWidget: property widget.
        """
        for row in range(self.__layout.rowCount()):
            item = self.__layout.itemAtPosition(row, 1)
            if item and name == item.widget().toolTip():
                return item.widget()


class NodePropWidget(QtWidgets.QWidget):
    """
    Node properties widget for display a Node object.

    Args:
        parent (QtWidgets.QWidget): parent object.
        node (NodeGraphQt.BaseNode): node.
    """

    #: signal (node_id, prop_name, prop_value)
    property_changed = QtCore.Signal(str, str, object)
    property_closed = QtCore.Signal(str)

    def __init__(self, parent=None, node=None):
        super(NodePropWidget, self).__init__(parent)
        self.__node_id = node.id
        self.__tab_windows = {}
        self.__tab = QtWidgets.QTabWidget()

        # close_btn = QtWidgets.QPushButton('X')
        # close_btn.setToolTip('close property')
        # close_btn.clicked.connect(self._on_close)

        self.name_wgt = PropLineEdit()
        self.name_wgt.setToolTip('name')
        self.name_wgt.set_value(node.name())
        self.name_wgt.value_changed.connect(self._on_property_changed)

        # self.type_wgt = QtWidgets.QLabel(node.type_)
        # self.type_wgt.setAlignment(QtCore.Qt.AlignRight)
        # self.type_wgt.setToolTip('type_')
        # font = self.type_wgt.font()
        # font.setPointSize(10)
        # self.type_wgt.setFont(font)

        name_layout = QtWidgets.QHBoxLayout()
        name_layout.setContentsMargins(0, 0, 0, 0)
        name_layout.addWidget(QtWidgets.QLabel('Name'))
        name_layout.addWidget(self.name_wgt)
        # name_layout.addWidget(close_btn)
        layout = QtWidgets.QVBoxLayout(self)
        layout.setSpacing(4)
        layout.addLayout(name_layout)
        layout.addWidget(self.__tab)
        # layout.addWidget(self.type_wgt)
        self._read_node(node)

    def __repr__(self):
        return '<{} object at {}>'.format(
            self.__class__.__name__, hex(id(self))
        )

    def _on_close(self):
        """
        called by the close button.
        """
        self.property_closed.emit(self.__node_id)

    def _on_property_changed(self, name, value):
        """
        slot function called when a property widget has changed.

        Args:
            name (str): property name.
            value (object): new value.
        """
        self.property_changed.emit(self.__node_id, name, value)

    def _read_node(self, node):
        """
        Populate widget from a node.

        Args:
            node (NodeGraphQt.BaseNode): node class.
        """
        model = node.model
        graph_model = node.graph.model

        common_props = graph_model.get_node_common_properties(node.type_)

        # sort tabs and properties.
        tab_mapping = defaultdict(list)
        for prop_name, prop_val in model.custom_properties.items():
            tab_name = model.get_tab_name(prop_name)
            tab_mapping[tab_name].append((prop_name, prop_val))

        # add tabs.
        for tab in sorted(tab_mapping.keys()):
            if tab != 'Node':
                self.add_tab(tab)

        # populate tab properties.
        for tab in sorted(tab_mapping.keys()):
            prop_window = self.__tab_windows[tab]
            for prop_name, value in tab_mapping[tab]:
                wid_type = model.get_widget_type(prop_name)
                if wid_type == 0:
                    continue

                _WidgetClass = WIDGET_MAP.get(wid_type)
                widget = _WidgetClass()
                if prop_name in common_props.keys():
                    if 'items' in common_props[prop_name].keys():
                        widget.set_items(common_props[prop_name]['items'])
                    if 'range' in common_props[prop_name].keys():
                        prop_range = common_props[prop_name]['range']
                        widget.set_min(prop_range[0])
                        widget.set_max(prop_range[1])

                prop_window.add_widget(prop_name, widget, value,
                                       prop_name.replace('_', ' '))
                widget.value_changed.connect(self._on_property_changed)

        # add "Node" tab properties.
        self.add_tab('Node')
        # GX
        default_props = ['color', 'text_color'] #, 'disabled', 'id']
        prop_window = self.__tab_windows['Node']
        for prop_name in default_props:
            wid_type = model.get_widget_type(prop_name)
            _WidgetClass = WIDGET_MAP.get(wid_type)

            widget = _WidgetClass()
            prop_window.add_widget(prop_name,
                                   widget,
                                   model.get_property(prop_name),
                                   prop_name.replace('_', ' '))

            widget.value_changed.connect(self._on_property_changed)

        # self.type_wgt.setText(model.get_property('type_'))

    def node_id(self):
        """
        Returns the node id linked to the widget.

        Returns:
            str: node id
        """
        return self.__node_id

    def add_widget(self, name, widget, tab='Properties'):
        """
        add new node property widget.

        Args:
            name (str): property name.
            widget (BaseProperty): property widget.
            tab (str): tab name.
        """
        if tab not in self._widgets.keys():
            tab = 'Properties'
        window = self.__tab_windows[tab]
        window.add_widget(name, widget)
        widget.value_changed.connect(self._on_property_changed)

    def add_tab(self, name):
        """
        add a new tab.

        Args:
            name (str): tab name.

        Returns:
            PropListWidget: tab child widget.
        """
        if name in self.__tab_windows.keys():
            raise AssertionError('Tab name {} already taken!'.format(name))
        self.__tab_windows[name] = PropListWidget(self)
        self.__tab.addTab(self.__tab_windows[name], name)
        return self.__tab_windows[name]

    def get_widget(self, name):
        """
        get property widget.

        Args:
            name (str): property name.

        Returns:
            QtWidgets.QWidget: property widget.
        """
        if name == 'name':
            return self.name_wgt
        for tab_name, prop_win in self.__tab_windows.items():
            widget = prop_win.get_widget(name)
            if widget:
                return widget


if __name__ == '__main__':
    import sys
    from NodeGraphQt import BaseNode, NodeGraph


    class TestNode(BaseNode):
        NODE_NAME = 'test node'

        def __init__(self):
            super(TestNode, self).__init__()
            self.create_property('label_test', 'Test Text',
                                 widget_type=NODE_PROP_QLABEL)
            self.create_property('line_edit', 'Test Text',
                                 widget_type=NODE_PROP_QLINEEDIT)
            self.create_property('color_picker', (0, 0, 255,200),
                                 widget_type=NODE_PROP_COLORPICKER)
            self.create_property('integer', 10,
                                 widget_type=NODE_PROP_QSPINBOX)
            self.create_property('list', 'item1',
                                 items=['item1', 'item2', 'item3'],
                                 widget_type=NODE_PROP_QCOMBO)
            self.create_property('range', 50,
                                 range=(45, 55),
                                 widget_type=NODE_PROP_SLIDER)
            self.create_property('text_edit', 'Test Text',
                                 widget_type=NODE_PROP_QTEXTEDIT,
                                 tab='text')


    def prop_changed(node_id, prop_name, prop_value):
        print('-' * 100)
        print(node_id, prop_name, prop_value)


    def prop_close(node_id):
        print('=' * 100)
        print(node_id)


    app = QtWidgets.QApplication(sys.argv)

    graph = NodeGraph()
    graph.register_node(TestNode)

    test_node = graph.create_node('nodeGraphQt.nodes.TestNode')

    node_prop = NodePropWidget(node=test_node)
    node_prop.property_changed.connect(prop_changed)
    node_prop.property_closed.connect(prop_close)
    node_prop.show()

    app.exec_()



import argparse
import os
import sys
from pathlib import Path
import depthai as dai
from typing import Tuple

def getRgbResolution(resolution: str):
    """
    Parses Color camera resolution based on the string
    """
    resolution = str(resolution).upper()
    if resolution == '3120' or resolution == '13MP':
        return dai.ColorCameraProperties.SensorResolution.THE_13_MP
    elif resolution == '3040' or resolution == '12MP':
        return dai.ColorCameraProperties.SensorResolution.THE_12_MP
    elif resolution == '2160' or resolution == '4K':
        return dai.ColorCameraProperties.SensorResolution.THE_4_K
    # elif resolution == '1920' or resolution == '5MP':
    #     return dai.ColorCameraProperties.SensorResolution.THE_5_MP
    elif resolution == '800' or resolution == '800P':
        return dai.ColorCameraProperties.SensorResolution.THE_800_P
    elif resolution == '720' or resolution == '720P':
        return dai.ColorCameraProperties.SensorResolution.THE_720_P
    else: # Default
        return dai.ColorCameraProperties.SensorResolution.THE_1080_P

def getMonoResolution(resolution: str):
    """
    Parses Mono camera resolution based on the string
    """
    resolution = str(resolution).upper()
    if resolution == '800' or resolution == '800P':
        return dai.MonoCameraProperties.SensorResolution.THE_800_P
    elif resolution == '720' or resolution == '720P':
        return dai.MonoCameraProperties.SensorResolution.THE_720_P
    elif resolution == '480' or resolution == '480P':
        return dai.MonoCameraProperties.SensorResolution.THE_480_P
    else: # Default
        return dai.MonoCameraProperties.SensorResolution.THE_400_P

folderPath = Path(os.path.abspath(sys.argv[0])).parent
try:
    import cv2
    _colorMaps = list(map(lambda name: name[len("COLORMAP_"):], filter(lambda name: name.startswith("COLORMAP_"), vars(cv2))))
except:
    _colorMaps = None
_streamChoices = ("nnInput", "color", "left", "right", "depth", "depthRaw", "disparity", "disparityColor", "rectifiedLeft", "rectifiedRight")
_openvinoVersions = [v.replace("VERSION_", "") for v in vars(dai.OpenVINO.Version) if v.startswith("VERSION_")]
_orientationChoices = list(filter(lambda var: var[0].isupper(), vars(dai.CameraImageOrientation)))

def _checkRange(minVal, maxVal):
    def checkFn(value):
        ivalue = int(value)
        if minVal <= ivalue <= maxVal:
            return ivalue
        else:
            raise argparse.ArgumentTypeError(
                "{} is an invalid int value, must be in range {}..{}".format(value, minVal, maxVal)
            )

    return checkFn

def _commaSeparated(default, cast=str):
    def _fun(option):
        optionList = option.split(",")
        if len(optionList) not in [1, 2]:
            raise argparse.ArgumentTypeError(
                "{0} format is invalid. See --help".format(option)
            )
        elif len(optionList) == 1:
            return optionList[0], cast(default)
        else:
            try:
                return optionList[0], cast(optionList[1])
            except ValueError:
                raise argparse.ArgumentTypeError(
                    "In option: {0} {1} is not in a correct format!".format(option, optionList[1])
                )
    return _fun

def _orientationCast(arg):
    if not hasattr(dai.CameraImageOrientation, arg):
        raise argparse.ArgumentTypeError("Invalid camera orientation specified: '{}'. Available: {}".format(arg, _orientationChoices))
    return getattr(dai.CameraImageOrientation, arg)

def _checkEnum(enum):
    def _fun(value: str):
        try:
            return getattr(enum, value.upper())
        except:
            choices = [f"'{str(item).split('.')[-1]}'" for name, item in vars(enum).items() if name.isupper()]
            raise argparse.ArgumentTypeError(
                "{} option wasn't found in {} options! Choices: {}".format(value, enum, ', '.join(choices))
            )

    return _fun


class ArgsManager():
    @staticmethod
    def parseArgs(parser: argparse.ArgumentParser = None):
        """
        Creates Argument parser for common OAK device configuration

        Args:
            parser (argparse.ArgumentParser, optional): Use an existing parser. By default it creates a new ArgumentParser.
        """
        if parser is None:
            parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('-cam', '--camera', choices=["left", "right", "color"], default="color", help="Use one of DepthAI cameras for inference (conflicts with -vid)")
        parser.add_argument('-vid', '--video', type=str, help="Path to video file (or YouTube link) to be used for inference (conflicts with -cam)")
        parser.add_argument('-dd', '--disableDepth', action="store_true", help="Disable depth information")
        parser.add_argument('-dnn', '--disableNeuralNetwork', action="store_true", help="Disable neural network inference")
        parser.add_argument('-cnnp', '--cnnPath', type=Path, help="Path to cnn model directory to be run")
        parser.add_argument("-cnn", "--cnnModel", default="mobilenet-ssd", type=str,
                            help="Cnn model to run on DepthAI")
        parser.add_argument('-sh', '--shaves', type=int, help="Number of MyriadX SHAVEs to use for neural network blob compilation")
        parser.add_argument('-cnnsize', '--cnnInputSize', default=None, type=str,
                            help="Neural network input dimensions, in \"WxH\" format, e.g. \"544x320\"")
        
        # Color/Mono cam related arguments
        parser.add_argument("-rgbr", "--rgbResolution", default='1080P', type=str,
                            help="RGB cam res height: (1920x)1080, (3840x)2160, (4056x)3040, (1280x)720, (1280x)800. Default: %(default)s")
        parser.add_argument("-rgbf", "--rgbFps", default=30.0, type=float,
                            help="RGB cam fps: max 118.0 for H:1080, max 42.0 for H:2160. Default: %(default)s")
        parser.add_argument("-monor", "--monoResolution", default='400P', type=str,
                            help="Mono cam res height: (1280x)720, (1280x)800 or (640x)400. Default: %(default)s")
        parser.add_argument("-monof", "--monoFps", default=30.0, type=float,
                            help="Mono cam fps: max 60.0 for H:720 or H:800, max 120.0 for H:400. Default: %(default)s")
        parser.add_argument('-fps', '--fps', type=float, help='Camera FPS applied to all sensors')

        # ColorCamera ISP values
        parser.add_argument('-isp', '--ispScale', type=_commaSeparated(None), help="Sets ColorCamera's ISP scale")
        parser.add_argument('-sharpness', '--sharpness', default=None, type=_checkRange(0,4),
                            help="Sets ColorCamera's sharpness")
        parser.add_argument('-lumaDenoise', '--lumaDenoise', default=None, type=_checkRange(0, 4),
                            help="Sets ColorCamera's Luma denoise")
        parser.add_argument('-chromaDenoise', '--chromaDenoise', default=None, type=_checkRange(0, 4),
                            help="Sets ColorCamera's Chroma denoise")

        # ColorCamera controls
        parser.add_argument('-manualFocus', '--manualFocus', default=None, type=_checkRange(0, 255),
                            help="Specify a Lens Position between 0 and 255 to use manual focus. Otherwise, auto-focus is used by default.")
        parser.add_argument('-afMode', '--afMode', default=None, type=_checkEnum(dai.CameraControl.AutoFocusMode),
                            help="Specify the Auto Focus mode for the ColorCamera. AUTO by default.")
        parser.add_argument('-awbMode', '--awbMode', default=None, type=_checkEnum(dai.CameraControl.AutoWhiteBalanceMode),
                            help="Specify the Auto White Balance mode for the ColorCamera. AUTO by default.")
        parser.add_argument('-sceneMode', '--sceneMode', default=None, type=_checkEnum(dai.CameraControl.SceneMode),
                            help="Specify the Scene mode for the ColorCamera. AUTO by default.")
        parser.add_argument('-abMode', '-antiBandingMode', '--antiBandingMode', default=None, type=_checkEnum(dai.CameraControl.AntiBandingMode),
                            help="Specify the Anti-Banding mode for the ColorCamera. AUTO by default.")
        parser.add_argument('-effectMode', '--effectMode', default=None, type=_checkEnum(dai.CameraControl.EffectMode),
                            help="Specify the Effect mode for the ColorCamera. AUTO by default.")

        parser.add_argument("--cameraControls", action="store_true", help="Show camera configuration options in GUI and control them using keyboard")
        parser.add_argument("--cameraExposure", type=_commaSeparated("all", int), nargs="+", help="Specify camera saturation")
        parser.add_argument("--cameraSensitivity", type=_commaSeparated("all", int), nargs="+", help="Specify camera sensitivity")
        parser.add_argument("--cameraSaturation", type=_commaSeparated("all", int), nargs="+", help="Specify image saturation")
        parser.add_argument("--cameraContrast", type=_commaSeparated("all", int), nargs="+", help="Specify image contrast")
        parser.add_argument("--cameraBrightness", type=_commaSeparated("all", int), nargs="+", help="Specify image brightness")
        parser.add_argument("--cameraSharpness", type=_commaSeparated("all", int), nargs="+", help="Specify image sharpness")

        
        # Depth related arguments
        parser.add_argument("-dct", "--disparityConfidenceThreshold", default=245, type=_checkRange(0, 255),
                            help="Disparity confidence threshold, used for depth measurement. Default: %(default)s")
        parser.add_argument("-lrct", "--lrcThreshold", default=4, type=_checkRange(0, 10),
                            help="Left right check threshold, used for depth measurement. Default: %(default)s")
        parser.add_argument("-sig", "--sigma", default=0, type=_checkRange(0, 250),
                            help="Sigma value for Bilateral Filter applied on depth. Default: %(default)s")
        parser.add_argument("-med", "--stereoMedianSize", default=5, type=int, choices=[0, 3, 5, 7],
                            help="Disparity / depth median filter kernel size (N x N) . 0 = filtering disabled. Default: %(default)s")
        parser.add_argument('-dlrc', '--disableStereoLrCheck', action="store_false", dest="stereoLrCheck",
                            help="Disable stereo 'Left-Right check' feature.")
        parser.add_argument('-ext', '--extendedDisparity', action="store_true",
                            help="Enable stereo 'Extended Disparity' feature.")
        parser.add_argument('-sub', '--subpixel', action="store_true",
                            help="Enable stereo 'Subpixel' feature.")
        parser.add_argument("-cm", "--colorMap", default="JET", choices=_colorMaps, help="Change color map used to apply colors to depth/disparity frames. Default: %(default)s")
        
        # Spatial image detection related arguments
        parser.add_argument("-maxd", "--maxDepth", default=10000, type=int,
                            help="Maximum depth distance for spatial coordinates in mm. Default: %(default)s")
        parser.add_argument("-mind", "--minDepth", default=100, type=int,
                            help="Minimum depth distance for spatial coordinates in mm. Default: %(default)s")
        parser.add_argument('-sbb', '--spatialBoundingBox', action="store_true",
                            help="Display spatial bounding box (ROI) when displaying spatial information. The Z coordinate get's calculated from the ROI (average)")
        parser.add_argument("-sbbsf", "--sbbScaleFactor", default=0.3, type=float,
                            help="Spatial bounding box scale factor. Sometimes lower scale factor can give better depth (Z) result. Default: %(default)s")
        
        parser.add_argument('-s', '--show', default=[], nargs="+", choices=_streamChoices, help="Choose which previews to show. Default: %(default)s")
        parser.add_argument("-dff", "--disableFullFovNn", default=False, action="store_true",
                            help="Disable full RGB FOV for NN, keeping the nn aspect ratio")
        parser.add_argument('--report', nargs="+", default=[], choices=["temp", "cpu", "memory"], help="Display device utilization data")
        parser.add_argument('--reportFile', help="Save report data to specified target file in CSV format")
        parser.add_argument('-cb', '--callback', type=Path, default=folderPath / 'callbacks.py', help="Path to callbacks file to be used. Default: %(default)s")
        parser.add_argument("--openvinoVersion", type=str, choices=_openvinoVersions, help="Specify which OpenVINO version to use in the pipeline")
        parser.add_argument("--count", type=str, dest='countLabel',
                            help="Count and display the number of specified objects on the frame. You can enter either the name of the object or its label id (number).")
        parser.add_argument("-dev", "--deviceId", type=str,
                            help="DepthAI MX id of the device to connect to. Use the word 'list' to show all devices and exit.")
        parser.add_argument('-bandw', '--bandwidth', type=str, default="auto", choices=["auto", "low", "high"], help="Force bandwidth mode. \n"
                                                                                                                    "If set to \"high\", the output streams will stay uncompressed\n"
                                                                                                                    "If set to \"low\", the output streams will be MJPEG-encoded\n"
                                                                                                                    "If set to \"auto\" (default), the optimal bandwidth will be selected based on your connection type and speed")
        parser.add_argument('-gt', '--guiType', type=str, default="auto", choices=["auto", "qt", "cv"], help="Specify GUI type of the demo. \"cv\" uses built-in OpenCV display methods, \"qt\" uses Qt to display interactive GUI. \"auto\" will use OpenCV for Raspberry Pi and Qt for other platforms")
        parser.add_argument('-usbs', '--usbSpeed', type=str, default="usb3", choices=["usb2", "usb3"], help="Force USB communication speed. Default: %(default)s")
        parser.add_argument('-enc', '--encode', type=_commaSeparated(default=30.0, cast=float), nargs="+", default=[],
                            help="Define which cameras to encode (record) \n"
                                "Format: cameraName or cameraName,encFps \n"
                                "Example: -enc left color \n"
                                "Example: -enc color right,10 left,10")
        parser.add_argument('-encout', '--encodeOutput', type=Path, default=folderPath / 'recordings', help="Path to directory where to store encoded files. Default: %(default)s")
        parser.add_argument('-xls', '--xlinkChunkSize', type=int, help="Specify XLink chunk size")
        parser.add_argument('-poeq', '--poeQuality', type=_checkRange(1, 100), default=100, help="Specify PoE encoding video quality (1-100)")
        parser.add_argument('-camo', '--cameraOrientation', type=_commaSeparated(default="AUTO", cast=_orientationCast), nargs="+", default=[],
                            help=("Define cameras orientation (available: {}) \n"
                                "Format: camera_name,camera_orientation \n"
                                "Example: -camo color,ROTATE_180_DEG right,ROTATE_180_DEG left,ROTATE_180_DEG").format(', '.join(_orientationChoices))
                            )
        parser.add_argument("--irDotBrightness", type=_checkRange(0, 1200), default=0, help="For OAK-D Pro: specify IR dot projector brightness, range: 0..1200 [mA], default 0 (turned off)")
        parser.add_argument("--irFloodBrightness", type=_checkRange(0, 1500), default=0, help="For OAK-D Pro: specify IR flood illumination brightness, range: 0..1500 [mA], default 0 (turned off)")
        parser.add_argument('--skipVersionCheck', action="store_true", help="Disable libraries version check")
        parser.add_argument('--noSupervisor', action="store_true", help="Disable supervisor check")
        parser.add_argument('--sync', action="store_true", help="Enable frame and NN synchronization. If enabled, all frames and NN results will be synced before preview (same sequence number)")
        parser.add_argument('--noRgbDepthAlign', action="store_true", help="Disable RGB-Depth align (depth frame will be aligned with the RGB frame)")
        parser.add_argument('--debug', action="store_true", help="Enables debug mode. Capability to connect to already BOOTED devices and also implicitly disables version check")
        parser.add_argument("-app","--app", type=str, choices=["uvc", "record"], help="Specify which app to run instead of the demo")
        parser.add_argument('-tun', '--cameraTuning', type=Path, help="Path to camera tuning blob to use, overriding the built-in tuning")
        
        args = parser.parse_args()
        # Parse arguments
        args.rgbResolution = getRgbResolution(args.rgbResolution)
        args.monoResolution = getMonoResolution(args.monoResolution)
        # Global FPS setting, applied to all cameras
        if args.fps is not None:
            args.rgbFps = args.fps
            args.monoFps = args.fps
        return args

    @staticmethod
    def parseApp() -> str:
        """
        Returns app name specified in the arguments, or None, if no app was specified.
        """
        parser = argparse.ArgumentParser(add_help=False)
        parser.add_argument("-app","--app", type=str, choices=["uvc", "record"], help="Specify which app to run instead of the demo")
        known, unknown = parser.parse_known_args()
        return known.app



    



import math
from enum import IntEnum
from types import SimpleNamespace
from typing import Tuple, Union, List, Any, Dict

import cv2
import depthai as dai
import distinctipy
import numpy as np

from depthai_sdk.classes.packets import (
    DetectionPacket,
    _TwoStageDetection,
    SpatialBbMappingPacket,
    TrackerPacket,
    _TrackingDetection
)
from depthai_sdk.oak_outputs.normalize_bb import NormalizeBoundingBox


class FramePosition(IntEnum):
    """
    Where on frame do we want to print text.
    """
    TopLeft = 0
    MidLeft = 1
    BottomLeft = 2
    TopMid = 10
    Mid = 11
    BottomMid = 12
    TopRight = 20
    MidRight = 21
    BottomRight = 22


default_color_map = cv2.applyColorMap(np.arange(256, dtype=np.uint8), cv2.COLORMAP_TURBO)
default_color_map[0] = [0, 0, 0]


class VisualizerHelper:
    bg_color = (0, 0, 0)
    front_color = (255, 255, 255)
    text_type = cv2.FONT_HERSHEY_SIMPLEX
    line_type = cv2.LINE_AA

    @classmethod
    def putText(cls,
                frame: np.ndarray,
                text: str,
                coords: Tuple[int, int],
                scale: float = 1.0,
                backColor: Tuple[int, int, int] = None,
                color: Tuple[int, int, int] = None):
        # Background text

        cv2.putText(frame, text, coords, cls.text_type, scale,
                    color=backColor if backColor else cls.bg_color,
                    thickness=int(scale * 3),
                    lineType=cls.line_type)
        # Front text
        cv2.putText(frame, text, coords, cls.text_type, scale,
                    color=(int(color[0]), int(color[1]), int(color[2])) if color else cls.front_color,
                    thickness=int(scale),
                    lineType=cls.line_type)

    @classmethod
    def line(cls,
             frame: np.ndarray,
             p1: Tuple[int, int], p2: Tuple[int, int],
             color=None,
             thickness: int = 1) -> None:
        cv2.line(frame, p1, p2,
                 cls.bg_color,
                 thickness * 3,
                 cls.line_type)
        cv2.line(frame, p1, p2,
                 (int(color[0]), int(color[1]), int(color[2])) if color else cls.front_color,
                 thickness,
                 cls.line_type)

    @classmethod
    def print_on_roi(cls, frame, topLeft, bottomRight, text: str, position: FramePosition = FramePosition.BottomLeft,
                     padPx=10):
        frame_roi = frame[topLeft[1]:bottomRight[1], topLeft[0]:bottomRight[0]]
        cls.print(frame=frame_roi, text=text, position=position, padPx=padPx)

    @classmethod
    def print(cls, frame, text: str, position: FramePosition = FramePosition.BottomLeft, padPx=10):
        """
        Prints text on the frame.
        @param frame: Frame
        @param text: Text to be printed
        @param position: Where on frame we want to print the text
        @param padPx: Padding (in pixels)
        """
        text_size = cv2.getTextSize(text, VisualizerHelper.text_type, fontScale=1.0, thickness=1)[0]
        frame_w = frame.shape[1]
        frame_h = frame.shape[0]

        y_pos = int(position) % 10
        if y_pos == 0:  # Y Top
            y = text_size[1] + padPx
        elif y_pos == 1:  # Y Mid
            y = int(frame_h / 2) + int(text_size[1] / 2)
        else:  # y_pos == 2. Y Bottom
            y = frame_h - padPx

        x_pos = int(position) // 10
        if x_pos == 0:  # X Left
            x = padPx
        elif x_pos == 1:  # X Mid
            x = int(frame_w / 2) - int(text_size[0] / 2)
        else:  # x_pos == 2  # X Right
            x = frame_w - text_size[0] - padPx
        cls.putText(frame, text, (x, y))


# def rectangle(frame, bbox, color: Tuple[int, int, int] = None):
#     x1, y1, x2, y2 = bbox
#     cv2.rectangle(frame, (x1, y1), (x2, y2), bg_color, 3)
#     cv2.rectangle(frame, pt1=(x1, y1), pt2=(x2, y2),
#                   color=(int(color[0]), int(color[1]), int(color[2])) if color else front_color, thickness=1)


def rectangle(src,
              bbox: np.ndarray,
              color: Tuple[int, int, int],
              thickness=-1,
              radius=0.1,
              line_type=cv2.LINE_AA,
              alpha=0.15):
    """
    Draw the rectangle (bounding box) on the frame.
    @param src: Frame
    @param top_left: Top left corner of the bounding box
    @param bottom_right: Bottom right corner of the bounding box
    @param color: Color of the rectangle
    @param thickness: Thickness of the rectangle. If -1, it will colorize the rectangle as well (with alpha)
    @param radius: Radius of the corners (for rounded rectangle)
    @param line_type: Line type for the rectangle
    @param alpha: Alpha for colorizing of the rectangle
    @return: Frame
    """
    top_left = (bbox[0], bbox[1])
    bottom_right = (bbox[2], bbox[3])
    top_right = (bottom_right[0], top_left[1])
    bottom_left = (top_left[0], bottom_right[1])

    height = abs(bottom_right[1] - top_left[1])

    if radius > 1:
        radius = 1

    corner_radius = int(radius * (height / 2))

    if thickness < 0:
        overlay = src.copy()

        # big rect
        top_left_main_rect = (int(top_left[0] + corner_radius), int(top_left[1]))
        bottom_right_main_rect = (int(bottom_right[0] - corner_radius), int(bottom_right[1]))

        top_left_rect_left = (top_left[0], top_left[1] + corner_radius)
        bottom_right_rect_left = (bottom_left[0] + corner_radius, bottom_left[1] - corner_radius)

        top_left_rect_right = (top_right[0] - corner_radius, top_right[1] + corner_radius)
        bottom_right_rect_right = (bottom_right[0], bottom_right[1] - corner_radius)

        all_rects = [
            [top_left_main_rect, bottom_right_main_rect],
            [top_left_rect_left, bottom_right_rect_left],
            [top_left_rect_right, bottom_right_rect_right]
        ]

        [cv2.rectangle(overlay, pt1=rect[0], pt2=rect[1], color=color, thickness=thickness) for rect in all_rects]

        cv2.ellipse(overlay, (top_left[0] + corner_radius, top_left[1] + corner_radius), (corner_radius, corner_radius),
                    180.0, 0,
                    90, color, thickness, line_type)
        cv2.ellipse(overlay, (top_right[0] - corner_radius, top_right[1] + corner_radius),
                    (corner_radius, corner_radius),
                    270.0, 0,
                    90, color, thickness, line_type)
        cv2.ellipse(overlay, (bottom_right[0] - corner_radius, bottom_right[1] - corner_radius),
                    (corner_radius, corner_radius), 0.0, 0, 90,
                    color, thickness, line_type)
        cv2.ellipse(overlay, (bottom_left[0] + corner_radius, bottom_left[1] - corner_radius),
                    (corner_radius, corner_radius), 90.0, 0,
                    90, color, thickness, line_type)

        cv2.ellipse(src, (top_left[0] + corner_radius, top_left[1] + corner_radius), (corner_radius, corner_radius),
                    180.0, 0, 90,
                    color, 2, line_type)
        cv2.ellipse(src, (top_right[0] - corner_radius, top_right[1] + corner_radius), (corner_radius, corner_radius),
                    270.0, 0, 90,
                    color, 2, line_type)
        cv2.ellipse(src, (bottom_right[0] - corner_radius, bottom_right[1] - corner_radius),
                    (corner_radius, corner_radius), 0.0, 0, 90,
                    color, 2, line_type)
        cv2.ellipse(src, (bottom_left[0] + corner_radius, bottom_left[1] - corner_radius),
                    (corner_radius, corner_radius),
                    90.0, 0, 90,
                    color, 2, line_type)

        cv2.addWeighted(overlay, alpha, src, 1 - alpha, 0, src)
    else:  # Don't fill the rectangle
        # draw straight lines
        cv2.line(src, (top_left[0] + corner_radius, top_left[1]), (top_right[0] - corner_radius, top_right[1]), color,
                 abs(thickness), line_type)
        cv2.line(src, (top_right[0], top_right[1] + corner_radius), (bottom_right[0], bottom_right[1] - corner_radius),
                 color, abs(thickness), line_type)
        cv2.line(src, (bottom_right[0] - corner_radius, bottom_left[1]),
                 (bottom_left[0] + corner_radius, bottom_right[1]),
                 color, abs(thickness), line_type)
        cv2.line(src, (bottom_left[0], bottom_left[1] - corner_radius), (top_left[0], top_left[1] + corner_radius),
                 color,
                 abs(thickness), line_type)

        # draw arcs
        cv2.ellipse(src, (top_left[0] + corner_radius, top_left[1] + corner_radius), (corner_radius, corner_radius),
                    180.0, 0, 90,
                    color, thickness, line_type)
        cv2.ellipse(src, (top_right[0] - corner_radius, top_right[1] + corner_radius), (corner_radius, corner_radius),
                    270.0, 0, 90,
                    color, thickness, line_type)
        cv2.ellipse(src, (bottom_right[0] - corner_radius, bottom_right[1] - corner_radius),
                    (corner_radius, corner_radius), 0.0, 0, 90,
                    color, thickness, line_type)
        cv2.ellipse(src, (bottom_left[0] + corner_radius, bottom_left[1] - corner_radius),
                    (corner_radius, corner_radius),
                    90.0, 0, 90,
                    color, thickness, line_type)

    return src


def get_text_color(background, threshold=0.6):
    bck = np.array(background) / 256
    clr = distinctipy.get_text_color((bck[2], bck[1], bck[0]), threshold)
    clr = distinctipy.get_rgb256(clr)
    return (clr[2], clr[1], clr[0])


def draw_mappings(packet: SpatialBbMappingPacket):
    dets = packet.spatials.detections
    for det in dets:
        roi = det.boundingBoxMapping.roi
        roi = roi.denormalize(packet.frame.shape[1], packet.frame.shape[0])
        top_left = roi.topLeft()
        bottom_right = roi.bottomRight()
        x_min = int(top_left.x)
        y_min = int(top_left.y)
        x_max = int(bottom_right.x)
        y_max = int(bottom_right.y)

        cv2.rectangle(packet.frame, (x_min, y_min), (x_max, y_max), VisualizerHelper.bg_color, 3)
        cv2.rectangle(packet.frame, (x_min, y_min), (x_max, y_max), VisualizerHelper.front_color, 1)


def spatials_text(spatials: dai.Point3f):
    return SimpleNamespace(
        x="X: " + ("{:.1f}m".format(spatials.x / 1000) if not math.isnan(spatials.x) else "--"),
        y="Y: " + ("{:.1f}m".format(spatials.y / 1000) if not math.isnan(spatials.y) else "--"),
        z="Z: " + ("{:.1f}m".format(spatials.z / 1000) if not math.isnan(spatials.z) else "--"),
    )


def draw_detections(packet: Union[DetectionPacket, _TwoStageDetection, TrackerPacket],
                    norm: NormalizeBoundingBox,
                    label_map: List[Tuple[str, Tuple]] = None):
    """
    Draw object detections to the frame.

    @param frame: np.ndarray frame
    @param dets: dai.ImgDetections
    @param norm: Object that handles normalization of the bounding box
    @param label_map: Label map for the detections
    """
    img_detections = []
    if isinstance(packet, TrackerPacket):
        img_detections = [t.srcImgDetection for t in packet.daiTracklets.tracklets]
    elif isinstance(packet, DetectionPacket):
        img_detections = [det for det in packet.img_detections.detections]

    for detection in img_detections:
        bbox = norm.normalize(packet.frame, (detection.xmin, detection.ymin, detection.xmax, detection.ymax))

        if label_map:
            txt, color = label_map[detection.label]
        else:
            txt = str(detection.label)
            color = VisualizerHelper.front_color

        VisualizerHelper.putText(packet.frame, txt, (bbox[0] + 5, bbox[1] + 25), scale=0.9)
        if packet._is_spatial_detection():
            point = packet._get_spatials(detection) \
                if isinstance(packet, TrackerPacket) \
                else detection.spatialCoordinates
            VisualizerHelper.putText(packet.frame, spatials_text(point).x, (bbox[0] + 5, bbox[1] + 50), scale=0.7)
            VisualizerHelper.putText(packet.frame, spatials_text(point).y, (bbox[0] + 5, bbox[1] + 75), scale=0.7)
            VisualizerHelper.putText(packet.frame, spatials_text(point).z, (bbox[0] + 5, bbox[1] + 100), scale=0.7)

        rectangle(packet.frame, bbox, color=color, thickness=1, radius=0)
        packet._add_detection(detection, bbox, txt, color)


def draw_tracklet_id(packet: TrackerPacket):
    for det in packet.detections:
        # centroid = det.centroid()
        VisualizerHelper.print_on_roi(packet.frame, det.top_left, det.bottom_right,
                                      f"Id: {str(det.tracklet.id)}",
                                      FramePosition.TopMid)


def draw_breadcrumb_trail(packets: List[TrackerPacket]):
    packet = packets[-1]  # Current packet

    dict_: Dict[str, List[_TrackingDetection]] = {}
    valid_ids = [t.id for t in packet.daiTracklets.tracklets]
    for idx in valid_ids:
        dict_[str(idx)] = []

    for packet in packets:
        for det in packet.detections:
            if det.tracklet.id in valid_ids:
                dict_[str(det.tracklet.id)].append(det)

    for idx, list_ in dict_.items():
        for i in range(len(list_) - 1):
            VisualizerHelper.line(packet.frame, list_[i].centroid(), list_[i + 1].centroid(), color=list_[i].color)


def colorize_depth(depth_frame: Union[dai.ImgFrame, Any], color_map=None):
    if isinstance(depth_frame, dai.ImgFrame):
        depth_frame = depth_frame.getFrame()

    depth_frame_color = cv2.normalize(depth_frame, None, 256, 0, cv2.NORM_INF, cv2.CV_8UC3)
    depth_frame_color = cv2.equalizeHist(depth_frame_color)

    return cv2.applyColorMap(depth_frame_color, color_map or default_color_map)


def colorize_disparity(frame: Union[dai.ImgFrame, Any], multiplier: float, color_map=None):
    if isinstance(frame, dai.ImgFrame):
        frame = frame.getFrame()

    frame = (frame * multiplier).astype(np.uint8)

    return cv2.applyColorMap(frame, color_map or default_color_map)


def draw_bb_mappings(depth_frame: Union[dai.ImgFrame, Any], bb_mappings: dai.SpatialLocationCalculatorConfig):
    depth_frame_color = colorize_depth(depth_frame)
    roi_datas = bb_mappings.getConfigData()
    for roi_data in roi_datas:
        roi = roi_data.roi
        roi = roi.denormalize(depth_frame_color.shape[1], depth_frame_color.shape[0])
        top_left = roi.topLeft()
        bottom_right = roi.bottomRight()
        xmin = int(top_left.x)
        ymin = int(top_left.y)
        xmax = int(bottom_right.x)
        ymax = int(bottom_right.y)

        rectangle(depth_frame_color, (xmin, ymin, xmax, ymax), (255, 255, 255), 1)


def calc_disp_multiplier(device: dai.Device, size: Tuple[int, int]) -> float:
    calib = device.readCalibration()
    baseline = calib.getBaselineDistance(useSpecTranslation=True) * 10  # mm
    intrinsics = calib.getCameraIntrinsics(dai.CameraBoardSocket.RIGHT, size)
    focal_length = intrinsics[0][0]
    return baseline * focal_length


def hex_to_bgr(hex: str) -> Tuple[int, int, int]:
    """
    "#ff1f00" (red) => (0, 31, 255)
    """
    value = hex.lstrip('#')
    return tuple(int(value[i:i + 2], 16) for i in (4, 2, 0))



import json
import warnings
from pathlib import Path
from typing import Callable, Union, List, Dict

import blobconverter

from depthai_sdk.classes.nn_config import Config
from depthai_sdk.components.camera_component import CameraComponent
from depthai_sdk.components.component import Component
from depthai_sdk.components.integrations.roboflow import RoboflowIntegration
from depthai_sdk.components.multi_stage_nn import MultiStageNN, MultiStageConfig
from depthai_sdk.components.nn_helper import *
from depthai_sdk.components.parser import *
from depthai_sdk.components.stereo_component import StereoComponent
from depthai_sdk.oak_outputs.xout import XoutNnResults, XoutTwoStage, XoutSpatialBbMappings, XoutFrames, XoutTracker
from depthai_sdk.oak_outputs.xout_base import StreamXout, XoutBase
from depthai_sdk.replay import Replay


class NNComponent(Component):
    # Public properties
    node: Union[
        None,
        dai.node.NeuralNetwork,
        dai.node.MobileNetDetectionNetwork,
        dai.node.MobileNetSpatialDetectionNetwork,
        dai.node.YoloDetectionNetwork,
        dai.node.YoloSpatialDetectionNetwork,
    ]
    tracker: dai.node.ObjectTracker
    image_manip: dai.node.ImageManip = None  # ImageManip used to resize the input to match the expected NN input size

    # Private properties
    _ar_resize_mode: AspectRatioResizeMode = AspectRatioResizeMode.LETTERBOX  # Default
    _input: Union[CameraComponent, 'NNComponent']  # Input to the NNComponent node passed on initialization
    _stream_input: dai.Node.Output  # Node Output that will be used as the input for this NNComponent

    _blob: dai.OpenVINO.Blob = None
    _forced_version: Optional[dai.OpenVINO.Version] = None  # Forced OpenVINO version
    _size: Tuple[int, int]  # Input size to the NN
    _args: Dict = None
    _config: Dict = None
    _node_type: dai.node = dai.node.NeuralNetwork  # Type of the node for `node`
    _roboflow: RoboflowIntegration = None

    _multi_stage_nn: MultiStageNN = None
    _multi_stage_config: MultiStageConfig = None

    x_in = None
    _input_queue = None

    _spatial: Union[None, bool, StereoComponent] = None
    _replay: Replay  # Replay module

    # For visualizer
    _labels: List = None  # obj detector labels
    _handler: Callable = None  # Custom model handler for decoding

    def __init__(self,
                 pipeline: dai.Pipeline,
                 model: Union[str, Path, Dict],  # str for SDK supported model or Path to custom model's json
                 input: Union[CameraComponent, 'NNComponent'],
                 nn_type: Optional[str] = None,  # Either 'yolo' or 'mobilenet'
                 decode_fn: Optional[Callable] = None,
                 tracker: bool = False,  # Enable object tracker - only for Object detection models
                 spatial: Union[None, bool, StereoComponent] = None,
                 replay: Optional[Replay] = None,
                 args: Dict = None  # User defined args
                 ) -> None:
        """
        Neural Network component abstracts:
         - DepthAI API nodes: NeuralNetwork, *DetectionNetwork, *SpatialDetectionNetwork, ObjectTracker
         - Downloading NN models (supported SDK NNs), parsing NN json configs and setting up the pipeline based on it
         - Decoding NN results
         - MultiStage pipelines - cropping high-res frames based on detections and use them for second NN inferencing

        Args:
            model (Union[str, Path, Dict]): str for SDK supported model / Path to blob or custom model's json
            input: (Union[Component, dai.Node.Output]): Input to the NN. If nn_component that is object detector, crop HQ frame at detections (Script node + ImageManip node)
            nn_type (str, optional): Type of the NN - Either 'Yolo' or 'MobileNet'
            tracker (bool, default False): Enable object tracker - only for Object detection models
            spatial (bool, default False): Enable getting Spatial coordinates (XYZ), only for Obj detectors. Yolo/SSD use on-device spatial calc, others on-host (gen2-calc-spatials-on-host)
            replay (Replay object): Replay
            args (Any, optional): Use user defined arguments when constructing the pipeline
        """
        super().__init__()
        self.out = self.Out(self)

        # Save passed settings
        self._input = input
        self._spatial = spatial
        self._args = args
        self._replay = replay
        self._decode_fn = decode_fn or None

        self.tracker = pipeline.createObjectTracker() if tracker else None

        # Parse passed settings
        self._parse_model(model)
        if nn_type:
            self._parse_node_type(nn_type)

        # Create NN node
        self.node = pipeline.create(self._node_type)

        if self._config and 'nn_config' in self._config:
            nn_config = self._config.get("nn_config", {})

            meta = nn_config.get('NN_specific_metadata', None)
            if self._is_yolo() and meta:
                self.config_yolo_from_metadata(metadata=meta)

    def _forced_openvino_version(self) -> dai.OpenVINO.Version:
        """
        Checks whether the component forces a specific OpenVINO version. This function is called after
        Camera has been configured and right before we connect to the OAK camera.
        @return: Forced OpenVINO version (optional).
        """
        return self._forced_version

    def _update_device_info(self, pipeline: dai.Pipeline, device: dai.Device, version: dai.OpenVINO.Version):

        if self._roboflow:
            path = self._roboflow.device_update(device)
            self._parse_config(path)

        if self._blob is None:
            self._blob = dai.OpenVINO.Blob(self._blob_from_config(self._config['model'], version))

        # TODO: update NN input based on camera resolution
        self.node.setBlob(self._blob)
        self._out = self.node.out

        if 1 < len(self._blob.networkInputs):
            raise NotImplementedError()

        nn_in: dai.TensorInfo = next(iter(self._blob.networkInputs.values()))
        # TODO: support models that expect mono img
        self._size: Tuple[int, int] = (nn_in.dims[0], nn_in.dims[1])
        # maxSize = dims

        if isinstance(self._input, CameraComponent):
            self._stream_input = self._input.stream
            self._setup_resize_manip(pipeline).link(self.node.input)
        elif self._is_multi_stage():
            # Calculate crop shape of the object detector
            frame_size = self._input._input.stream_size
            nn_size = self._input._size
            scale = frame_size[0] / nn_size[0], frame_size[1] / nn_size[1]
            i = 0 if scale[0] < scale[1] else 1
            crop = int(scale[i] * nn_size[0]), int(scale[i] * nn_size[1])
            # Crop the high-resolution frames, so it matches object detection frame aspect ratio

            self.image_manip = pipeline.createImageManip()
            self.image_manip.setNumFramesPool(10)
            self.image_manip_config = dai.ImageManipConfig()
            self.image_manip.initialConfig.setFrameType(dai.RawImgFrame.Type.BGR888p)

            self._input._stream_input.link(self.image_manip.inputImage)
            if self._input._is_detector() and self._decode_fn is None:
                self.image_manip.setResize(*crop)
                self.image_manip.setMaxOutputFrameSize(crop[0] * crop[1] * 3)

                # Create script node, get HQ frames from input.
                self._multi_stage_nn = MultiStageNN(pipeline, self._input.node, self.image_manip.out, self._size)
                self._multi_stage_nn.configure(self._multi_stage_config)
                self._multi_stage_nn.out.link(self.node.input)  # Cropped frames

                # For debugging, for integral counter
                self.node.out.link(self._multi_stage_nn.script.inputs['recognition'])
                self.node.input.setBlocking(True)
                self.node.input.setQueueSize(20)
            else:
                print('Using on-host decoding for multi-stage NN')
                # Custom NN
                self.image_manip.setResize(*self._size)
                self.image_manip.setMaxOutputFrameSize(self._size[0] * self._size[1] * 3)

                # TODO pass frame on device, and just send config from host
                self.x_in = pipeline.createXLinkIn()
                self.x_in.setStreamName("input_queue")
                self.x_in.setMaxDataSize(frame_size[0] * frame_size[1] * 3)
                self.x_in.out.link(self.image_manip.inputImage)

                self.x_in_cfg = pipeline.createXLinkIn()
                self.x_in_cfg.setStreamName("input_queue_cfg")
                self.x_in_cfg.out.link(self.image_manip.inputConfig)

                self.image_manip.out.link(self.node.input)
                self.node.input.setQueueSize(20)
        else:
            raise ValueError(
                "'input' argument passed on init isn't supported!"
                "You can only use NnComponent or CameraComponent as the input."
            )

        if self._spatial:
            if isinstance(self._spatial, bool):  # Create new StereoComponent
                self._spatial = StereoComponent(pipeline, args=self._args, replay=self._replay)
                self._spatial._update_device_info(pipeline, device, version)
            if isinstance(self._spatial, StereoComponent):
                self._spatial.depth.link(self.node.inputDepth)
                self._spatial.config_stereo(align=self._input._source)
            # Configure Spatial Detection Network

        if self._args:
            if self._is_spatial():
                self._config_spatials_args(self._args)

    def _parse_model(self, model):
        """
        Called when NNComponent is initialized. Parses "model" argument passed by user.
        """
        if isinstance(model, Dict):
            self._parse_config(model)
            return
        # Parse the input config/model
        elif isinstance(model, str):
            # Download from the web, or convert to Path
            model = getBlob(model) if isUrl(model) else Path(model)

        if model.suffix in ['.blob', '.json']:
            if model.suffix == '.blob':
                self._blob = dai.OpenVINO.Blob(model.resolve())
                self._forced_version = self._blob.version
            elif model.suffix == '.json':  # json config file was passed
                self._parse_config(model)
        else:  # SDK supported model
            models = getSupportedModels(printModels=False)
            zoo_models = blobconverter.zoo_list()

            if str(model) in models:
                model = models[str(model)] / 'config.json'
                self._parse_config(model)
            elif str(model) in zoo_models:
                print(
                    'Models from the OpenVINO Model Zoo do not carry any metadata'
                    ' (e.g., label map, decoding logic). Please keep this in mind when using models from Zoo.'
                )
                self._blob = dai.OpenVINO.Blob(blobconverter.from_zoo(str(model), shaves=6))
                self._forced_version = self._blob.version
            else:
                raise ValueError(f"Specified model '{str(model)}' is not supported by DepthAI SDK.\n"
                                 "Check SDK documentation page to see which models are supported.")

    def _parse_node_type(self, nnType: str) -> None:
        self._node_type = dai.node.NeuralNetwork
        if nnType:
            if nnType.upper() == 'YOLO':
                self._node_type = dai.node.YoloSpatialDetectionNetwork if self._is_spatial() else dai.node.YoloDetectionNetwork
            elif nnType.upper() == 'MOBILENET':
                self._node_type = dai.node.MobileNetSpatialDetectionNetwork if self._is_spatial() else dai.node.MobileNetDetectionNetwork

    def _config_spatials_args(self, args):
        if not isinstance(args, Dict):
            args = vars(args)  # Namespace -> Dict
        self.config_spatial(
            bb_scale_factor=args.get('sbbScaleFactor', None),
            lower_threshold=args.get('minDepth', None),
            upper_threshold=args.get('maxDepth', None),
        )

    def _parse_config(self, model_config: Union[Path, str, Dict]):
        """
        Called when NNComponent is initialized. Reads config.json file and parses relevant setting from there
        """
        parent_folder = None
        if isinstance(model_config, str):
            model_config = Path(model_config).resolve()

        if isinstance(model_config, Path):
            parent_folder = model_config.parent
            with model_config.open() as f:
                self._config = Config().load(json.loads(f.read()))
        else:  # Dict
            self._config = model_config

        if 'source' in self._config:
            if self._config['source'] == 'roboflow':
                from depthai_sdk.components.integrations.roboflow import RoboflowIntegration
                self._roboflow = RoboflowIntegration(self._config)
                self._parse_node_type('YOLO')  # Roboflow only supports YOLO models
                return
            else:
                raise ValueError(f"[NN Dict configuration] Source '{self._config['source']}' not supported")

        # Get blob from the config file
        if 'model' in self._config:
            model = self._config['model']

            # Resolve the paths inside config
            if parent_folder:
                for name in ['blob', 'xml', 'bin']:
                    if name in model:
                        model[name] = str((parent_folder / model[name]).resolve())

            if 'blob' in model:
                self._blob = dai.OpenVINO.Blob(model['blob'])

        # Parse OpenVINO version
        if "openvino_version" in self._config:
            self._forced_version = parse_open_vino_version(self._config.get("openvino_version"))

        # Save for visualization
        self._labels = self._config.get("mappings", {}).get("labels", None)

        # Handler.py logic to decode raw NN results into standardized AI results
        if 'handler' in self._config:
            self._handler = loadModule(model_config.parent / self._config["handler"])

            if not callable(getattr(self._handler, "decode", None)):
                raise RuntimeError("Custom model handler does not contain 'decode' method!")

        if 'nn_config' in self._config:
            nn_config = self._config.get("nn_config", {})

            # Parse node type
            nn_family = nn_config.get("NN_family", None)
            if nn_family:
                self._parse_node_type(nn_family)

    def _blob_from_config(self, model: Dict, version: dai.OpenVINO.Version) -> str:
        """
        Gets the blob from the config file.
        """
        vals = str(version).split('_')
        version_str = f"{vals[1]}.{vals[2]}"

        if 'model_name' in model:  # Use blobconverter to download the model
            zoo_type = model.get("zoo", 'intel')
            return blobconverter.from_zoo(model['model_name'],
                                          zoo_type=zoo_type,
                                          shaves=6,  # TODO: Calculate ideal shave amount
                                          version=version_str
                                          )

        if 'xml' in model and 'bin' in model:
            return blobconverter.from_openvino(xml=model['xml'],
                                               bin=model['bin'],
                                               data_type="FP16",  # Myriad X
                                               shaves=6,  # TODO: Calculate ideal shave amount
                                               version=version_str
                                               )

        raise ValueError("Specified `model` values in json config files are incorrect!")

    def _setup_resize_manip(self, pipeline: Optional[dai.Pipeline] = None) -> dai.Node.Output:
        """
        Creates ImageManip node that resizes the input to match the expected NN input size.
        DepthAI uses CHW (Planar) channel layout and BGR color order convention.
        """
        if not self.image_manip:
            self.image_manip = pipeline.create(dai.node.ImageManip)
            self._stream_input.link(self.image_manip.inputImage)
            self.image_manip.setMaxOutputFrameSize(self._size[0] * self._size[1] * 3)
            self.image_manip.initialConfig.setFrameType(dai.RawImgFrame.Type.BGR888p)
            # Set to non-blocking
            self.image_manip.inputImage.setBlocking(False)
            self.image_manip.inputImage.setQueueSize(2)

        # Set Aspect Ratio resizing mode
        if self._ar_resize_mode == AspectRatioResizeMode.CROP:
            # Cropping is already the default mode of the ImageManip node
            self.image_manip.initialConfig.setResize(self._size)
        elif self._ar_resize_mode == AspectRatioResizeMode.LETTERBOX:
            self.image_manip.initialConfig.setResizeThumbnail(*self._size)
        elif self._ar_resize_mode == AspectRatioResizeMode.STRETCH:
            self.image_manip.initialConfig.setResize(self._size)
            self.image_manip.setKeepAspectRatio(False)  # Not keeping aspect ratio -> stretching the image

        return self.image_manip.out

    def config_multistage_nn(self,
                             debug=False,
                             labels: Optional[List[int]] = None,
                             scale_bb: Optional[Tuple[int, int]] = None,
                             ) -> None:
        """
        Configures the MultiStage NN pipeline. Available if the input to this NNComponent is Detection NNComponent.

        Args:
            debug (bool, default False): Debug script node
            labels (List[int], optional): Crop & run inference only on objects with these labels
            scale_bb (Tuple[int, int], optional): Scale detection bounding boxes (x, y) before cropping the frame. In %.
        """
        if not self._is_multi_stage():
            print("Input to this model was not a NNComponent, so 2-stage NN inferencing isn't possible!"
                  "This configuration attempt will be ignored.")
            return

        self._multi_stage_config = MultiStageConfig(debug, labels, scale_bb)

    def _parse_label(self, label: Union[str, int]) -> int:
        if isinstance(label, int):
            return label
        elif isinstance(label, str):
            if not self._labels:
                raise ValueError("Incorrect trackLabels type! Make sure to pass NN configuration to"
                                 "the NNComponent so it can deccode string labels!")
            # Label map is Dict of either "name", or ["name", "color"]
            label_strs = [l.upper() if isinstance(l, str) else l[0].upper() for l in self._labels]

            if label.upper() not in label_strs: raise ValueError(f"String '{label}' wasn't found in passed labels!")
            return label_strs.index(label.upper())
        else:
            raise Exception('_parse_label only accepts int or str')

    def config_tracker(self,
                       tracker_type: Optional[dai.TrackerType] = None,
                       track_labels: Optional[List[int]] = None,
                       assignment_policy: Optional[dai.TrackerIdAssignmentPolicy] = None,
                       max_obj: Optional[int] = None,
                       threshold: Optional[float] = None
                       ):
        """
        Configure Object Tracker node (if it's enabled).

        Args:
            tracker_type (dai.TrackerType, optional): Set object tracker type
            track_labels (List[int], optional): Set detection labels to track
            assignment_policy (dai.TrackerType, optional): Set object tracker ID assignment policy
            max_obj (int, optional): Set max objects to track. Max 60.
            threshold (float, optional): Set threshold for object detection confidence. Default: 0.0
        """

        if self.tracker is None:
            warnings.warn("Tracker was not enabled! Enable with cam.create_nn('[model]', tracker=True)."
                          "This configuration attempt will be ignored.")
            return

        if tracker_type:
            self.tracker.setTrackerType(type=tracker_type)

        if track_labels and 0 < len(track_labels):
            labels = [self._parse_label(l) for l in track_labels]
            self.tracker.setDetectionLabelsToTrack(labels)

        if assignment_policy:
            self.tracker.setTrackerIdAssignmentPolicy(assignment_policy)

        if max_obj:
            if 60 < max_obj:
                raise ValueError("Maximum objects to track is 60!")
            self.tracker.setMaxObjectsToTrack(max_obj)

        if threshold:
            self.tracker.setTrackerThreshold(threshold)

    def config_yolo_from_metadata(self, metadata: Dict):
        """
        Configures (Spatial) Yolo Detection Network node with a dictionary. Calls config_yolo().
        """
        return self.config_yolo(
            num_classes=metadata['classes'],
            coordinate_size=metadata['coordinates'],
            anchors=metadata['anchors'],
            masks=metadata['anchor_masks'],
            iou_threshold=metadata['iou_threshold'],
            conf_threshold=metadata['confidence_threshold'],
        )

    def config_yolo(self,
                    num_classes: int,
                    coordinate_size: int,
                    anchors: List[float],
                    masks: Dict[str, List[int]],
                    iou_threshold: float,
                    conf_threshold: Optional[float] = None,
                    ) -> None:
        """
        Configures (Spatial) Yolo Detection Network node.

        """
        if not self._is_yolo():
            print('This is not a YOLO detection network! This configuration attempt will be ignored.')
            return

        self.node.setNumClasses(num_classes)
        self.node.setCoordinateSize(coordinate_size)
        self.node.setAnchors(anchors)
        self.node.setAnchorMasks(masks)
        self.node.setIouThreshold(iou_threshold)

        if conf_threshold:
            self.node.setConfidenceThreshold(conf_threshold)

    def config_nn(self,
                  conf_threshold: Optional[float] = None,
                  aspect_ratio_resize_mode: Union[AspectRatioResizeMode, str] = None):
        """
        Configures the Detection Network node.

        Args:
            conf_threshold: (float, optional): Confidence threshold for the detections (0..1]
            aspect_ratio_resize_mode: (AspectRatioResizeMode, optional): Change aspect ratio resizing mode - to either STRETCH, CROP, or LETTERBOX.
        """
        if aspect_ratio_resize_mode:
            if isinstance(aspect_ratio_resize_mode, str):
                try:
                    aspect_ratio_resize_mode = AspectRatioResizeMode[aspect_ratio_resize_mode.upper()]
                except (AttributeError, KeyError):
                    print('AR resize mode was not recognizied.'
                          'Options (case insensitive): STRETCH, CROP, LETTERBOX.'
                          'Using default LETTERBOX mode.')

            self._ar_resize_mode = aspect_ratio_resize_mode
        if conf_threshold and self._is_detector():
            self.node.setConfidenceThreshold(conf_threshold)

    def config_spatial(self,
                       bb_scale_factor: Optional[float] = None,
                       lower_threshold: Optional[int] = None,
                       upper_threshold: Optional[int] = None,
                       calc_algo: Optional[dai.SpatialLocationCalculatorAlgorithm] = None):
        """
        Configures the Spatial Detection Network node.

        Args:
            bb_scale_factor (float, optional): Specifies scale factor for detected bounding boxes (0..1]
            lower_threshold (int, optional): Specifies lower threshold in depth units (millimeter by default) for depth values which will used to calculate spatial data
            upper_threshold (int, optional): Specifies upper threshold in depth units (millimeter by default) for depth values which will used to calculate spatial data
            calc_algo (dai.SpatialLocationCalculatorAlgorithm, optional): Specifies spatial location calculator algorithm: Average/Min/Max
        """
        if not self._is_spatial():
            print('This is not a Spatial Detection network! This configuration attempt will be ignored.')
            return

        if bb_scale_factor:
            self.node.setBoundingBoxScaleFactor(bb_scale_factor)
        if lower_threshold:
            self.node.setDepthLowerThreshold(lower_threshold)
        if upper_threshold:
            self.node.setDepthUpperThreshold(upper_threshold)
        if calc_algo:
            self.node.setSpatialCalculationAlgorithm(calc_algo)

    """
    Available outputs (to the host) of this component
    """

    class Out:
        _comp: 'NNComponent'

        def __init__(self, nn_component: 'NNComponent'):
            self._comp = nn_component

        def main(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutBase:
            """
            Default output. Streams NN results and high-res frames that were downscaled and used for inferencing.
            Produces DetectionPacket or TwoStagePacket (if it's 2. stage NNComponent).
            """

            if self._comp._is_multi_stage():
                out = XoutTwoStage(det_nn=self._comp._input,
                                   second_nn=self._comp,
                                   frames=self._comp._input._input.get_stream_xout(),
                                   det_out=StreamXout(self._comp._input.node.id, self._comp._input.node.out),
                                   second_nn_out=StreamXout(self._comp.node.id, self._comp.node.out),
                                   device=device,
                                   input_queue_name="input_queue" if self._comp.x_in else None)
            else:
                out = XoutNnResults(det_nn=self._comp,
                                    frames=self._comp._input.get_stream_xout(),
                                    nn_results=StreamXout(self._comp.node.id, self._comp.node.out))

            return self._comp._create_xout(pipeline, out)

        def passthrough(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutBase:
            """
            Default output. Streams NN results and passthrough frames (frames used for inferencing)
            Produces DetectionPacket or TwoStagePacket (if it's 2. stage NNComponent).
            """
            if self._comp._is_multi_stage():
                out = XoutTwoStage(det_nn=self._comp._input,
                                   second_nn=self._comp,
                                   frames=StreamXout(self._comp._input.node.id, self._comp._input.node.passthrough),
                                   det_out=StreamXout(self._comp._input.node.id, self._comp._input.node.out),
                                   second_nn_out=StreamXout(self._comp.node.id, self._comp.node.out),
                                   device=device,
                                   input_queue_name="input_queue" if self._comp.x_in else None)
            else:
                out = XoutNnResults(det_nn=self._comp,
                                    frames=StreamXout(self._comp.node.id, self._comp.node.passthrough),
                                    nn_results=StreamXout(self._comp.node.id, self._comp.node.out))

            return self._comp._create_xout(pipeline, out)

        def image_manip(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutBase:
            out = XoutFrames(frames=StreamXout(self._comp.image_manip.id, self._comp.image_manip.out))
            return self._comp._create_xout(pipeline, out)

        def input(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutBase:
            out = XoutFrames(frames=StreamXout(self._comp._input.node.id, self._comp._stream_input))
            return self._comp._create_xout(pipeline, out)

        def spatials(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutSpatialBbMappings:
            """
            Streams depth and bounding box mappings (``SpatialDetectionNework.boundingBoxMapping``). Produces SpatialBbMappingPacket.
            """
            if not self._comp._is_spatial():
                raise Exception('SDK tried to output spatial data (depth + bounding box mappings),'
                                'but this is not a Spatial Detection network!')

            out = XoutSpatialBbMappings(device=device,
                                        frames=StreamXout(self._comp.node.id, self._comp.node.passthroughDepth),
                                        configs=StreamXout(self._comp.node.id, self._comp.node.out))

            return self._comp._create_xout(pipeline, out)

        def twostage_crops(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutFrames:
            """
            Streams 2. stage cropped frames to the host. Produces FramePacket.
            """
            if not self._comp._is_multi_stage():
                raise Exception('SDK tried to output TwoStage crop frames, but this is not a Two-Stage NN component!')

            out = XoutFrames(
                frames=StreamXout(self._comp._multi_stage_nn.manip.id, self._comp._multi_stage_nn.manip.out)
            )

            return self._comp._create_xout(pipeline, out)

        def tracker(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutTracker:
            """
            Streams ObjectTracker tracklets and high-res frames that were downscaled and used for inferencing. Produces TrackerPacket.
            """
            if not self._comp._is_tracker():
                raise Exception('Tracker was not enabled! Enable with cam.create_nn("[model]", tracker=True)!')

            self._comp.node.passthrough.link(self._comp.tracker.inputDetectionFrame)
            self._comp.node.out.link(self._comp.tracker.inputDetections)

            # TODO: add support for full frame tracking
            self._comp.node.passthrough.link(self._comp.tracker.inputTrackerFrame)

            out = XoutTracker(self._comp,
                              self._comp._input.get_stream_xout(),  # CameraComponent
                              StreamXout(self._comp.tracker.id, self._comp.tracker.out)
                              )
            return self._comp._create_xout(pipeline, out)

    out: Out

    # Checks
    def _is_spatial(self) -> bool:
        return self._spatial is not None  # todo fix if spatial is bool and equals to False

    def _is_tracker(self) -> bool:
        # Currently, only object detectors are supported
        return self._is_detector() and self.tracker is not None

    def _is_yolo(self) -> bool:
        return (
                self._node_type == dai.node.YoloDetectionNetwork or
                self._node_type == dai.node.YoloSpatialDetectionNetwork
        )

    def _is_mobile_net(self) -> bool:
        return (
                self._node_type == dai.node.MobileNetDetectionNetwork or
                self._node_type == dai.node.MobileNetSpatialDetectionNetwork
        )

    def _is_detector(self) -> bool:
        """
        Currently these 2 object detectors are supported
        """
        return self._is_yolo() or self._is_mobile_net()

    def _is_multi_stage(self):
        if not isinstance(self._input, type(self)):
            return False

        if not isinstance(self._input, Component):
            return False

        # if not self._input._is_detector():
        #     raise Exception('Only object detector models can be used as an input to the NNComponent!')

        return True



from typing import Tuple

import numpy as np

from depthai_sdk.components.nn_helper import AspectRatioResizeMode


class NormalizeBoundingBox:
    """
    Normalized bounding box (BB) received from the device. It will also take into account type of aspect ratio
    resize mode and map coordinates to correct location.
    """

    def __init__(self, aspect_ratio: Tuple[float, float], ar_resize_mode: AspectRatioResizeMode):
        """
        :param aspect_ratio: NN input size
        :param ar_resize_mode: Aspect ratio resize mode
        """
        self.aspect_ratio = aspect_ratio
        self.ar_resize_mode = ar_resize_mode

    def normalize(self, frame, bbox: Tuple[float, float, float, float]) -> np.ndarray:
        """
        Mapps bounding box coordinates (0..1) to pixel values on frame

        Args:
            frame (numpy.ndarray): Frame to which adjust the bounding box
            bbox (list): list of bounding box points in a form of :code:`[x1, y1, x2, y2, ...]`

        Returns:
            list: Bounding box points mapped to pixel values on frame
        """
        bbox = np.array(bbox)

        # Edit the bounding boxes before normalizing them
        if self.ar_resize_mode == AspectRatioResizeMode.CROP:
            ar_diff = (self.aspect_ratio[0] / self.aspect_ratio[1]) / (frame.shape[1] / frame.shape[0])
            if ar_diff < 1:
                new_w = frame.shape[1] * ar_diff
                new_h = frame.shape[0]
                bbox[0] = bbox[0] * new_w + (frame.shape[1] - new_w) / 2
                bbox[1] = bbox[1] * new_h
                bbox[2] = bbox[2] * new_w + (frame.shape[1] - new_w) / 2
                bbox[3] = bbox[3] * new_h
            else:
                new_w = frame.shape[1]
                new_h = frame.shape[0] / ar_diff
                bbox[0] = bbox[0] * new_w
                bbox[1] = bbox[1] * new_h + (new_h - frame.shape[0]) / 2
                bbox[2] = bbox[2] * new_w
                bbox[3] = bbox[3] * new_h + (new_h - frame.shape[0]) / 2

            return bbox.astype(int)
        elif self.ar_resize_mode == AspectRatioResizeMode.STRETCH:
            # No need to edit bounding boxes when stretching
            pass
        elif self.ar_resize_mode == AspectRatioResizeMode.LETTERBOX:
            # There might be better way of doing this. TODO: test if it works as expected
            ar_diff = self.aspect_ratio[0] / self.aspect_ratio[1] - frame.shape[1] / frame.shape[0]
            sel = 0 if 0 < ar_diff else 1
            nsel = 0 if sel == 1 else 1
            # Get the divisor
            div = frame.shape[sel] / self.aspect_ratio[nsel]
            letterboxing_ratio = 1 - (frame.shape[nsel] / div) / self.aspect_ratio[sel]

            bbox[sel::2] -= abs(letterboxing_ratio) / 2
            bbox[sel::2] /= 1 - abs(letterboxing_ratio)

        # Normalize bounding boxes
        normVals = np.full(len(bbox), frame.shape[0])
        normVals[::2] = frame.shape[1]
        return (np.clip(bbox, 0, 1) * normVals).astype(int)



#!/usr/bin/python
# -*- coding: utf-8 -*-


class NodeMenuError(Exception): pass


class NodePropertyError(Exception): pass


class NodeWidgetError(Exception): pass


class NodeRegistrationError(Exception): pass


class PortError(Exception): pass


class PortRegistrationError(Exception): pass



from typing import List, Dict, Any

import depthai as dai

from depthai_sdk.oak_outputs.xout_base import XoutBase


class OakDevice:
    device: dai.Device = None
    # fpsHandlers: Dict[str, FPS] = dict()
    oak_out_streams: List[XoutBase] = []

    @property
    def image_sensors(self) -> List[dai.CameraBoardSocket]:
        """
        Available imageSensors available on the camera
        """
        return self.device.getConnectedCameras()

    @property
    def info(self) -> dai.DeviceInfo:
        return self.device.getDeviceInfo()

    def init_callbacks(self, pipeline: dai.Pipeline):
        for node in pipeline.getAllNodes():
            if isinstance(node, dai.node.XLinkOut):
                stream_name = node.getStreamName()
                # self.fpsHandlers[name] = FPS()
                self.device.getOutputQueue(stream_name, maxSize=4, blocking=False).addCallback(
                    lambda name, msg: self.new_msg(name, msg)
                )

    def new_msg(self, name, msg):
        for sync in self.oak_out_streams:
            sync.new_msg(name, msg)

    def check_sync(self):
        """
        Checks whether there are new synced messages, non-blocking.
        """
        for sync in self.oak_out_streams:
            sync.check_queue(block=False)  # Don't block!



import time
msgs = dict()
cntr = 0
def add_msg(msg, name, seq = None):
    global msgs
    if seq is None:
        seq = msg.getSequenceNum()
    seq = str(seq)
    ${DEBUG}node.warn(f"New msg {name}, seq {seq}")

    # Each seq number has its own dict of msgs
    if seq not in msgs:
        msgs[seq] = dict()
    msgs[seq][name] = msg

    # To avoid freezing
    if 15 < len(msgs):
        ${DEBUG}node.warn(f"Removing first element! len {len(msgs)}")
        msgs.popitem() # Remove first element

def get_msgs():
    global msgs
    seq_remove = [] # Arr of sequence numbers to get deleted
    for seq, syncMsgs in msgs.items():
        seq_remove.append(seq) # Will get removed from dict if we find synced msgs pair
        # Check if we have both detections and color frame with this sequence number
        if len(syncMsgs) == 2: # 1 frame, 1 detection
            for rm in seq_remove:
                del msgs[rm]
            ${DEBUG}node.warn(f"synced {seq}. Removed older sync values. len {len(msgs)}")
            return syncMsgs # Returned synced msgs
    return None

def correct_bb(xmin,ymin,xmax,ymax):
    if xmin < 0: xmin = 0.001
    if ymin < 0: ymin = 0.001
    if xmax > 1: xmax = 0.999
    if ymax > 1: ymax = 0.999
    return [xmin,ymin,xmax,ymax]

while True:
    time.sleep(0.001) # Avoid lazy looping

    frame = node.io['frames'].tryGet()
    if frame is not None:
        add_msg(frame, 'frames')

    dets = node.io['detections'].tryGet()
    if dets is not None:
        add_msg(dets, 'detections')

    rec = node.io['recognition'].tryGet()
    if rec is not None:
        cntr -= 1
        ${DEBUG}node.warn(f"Recognition results received. cntr {cntr}")

    sync_msgs = get_msgs()
    if sync_msgs is not None:
        img = sync_msgs['frames']
        dets = sync_msgs['detections']
        if 10 < cntr:
            ${DEBUG}node.warn(f"NN too slow, skipping frame. Cntr {cntr}")
            continue # If recognition model is too slow
        for i, det in enumerate(dets.detections):
            ${CHECK_LABELS}
            cfg = ImageManipConfig()
            bb = correct_bb(
                det.xmin${SCALE_BB_XMIN},
                det.ymin${SCALE_BB_YMIN},
                det.xmax${SCALE_BB_XMAX},
                det.ymax${SCALE_BB_YMAX},
                )
            cfg.setCropRect(*bb)
            ${DEBUG}node.warn(f"Sending {i + 1}. det. Det {det.xmin}, {det.ymin}, {det.xmax}, {det.ymax}")
            cfg.setResize(${WIDTH}, ${HEIGHT})
            cfg.setKeepAspectRatio(False)
            node.io['manip_cfg'].send(cfg)
            node.io['manip_img'].send(img)
            cntr += 1
            ${DEBUG}node.warn(f"Frame sent to recognition NN. cntr {cntr}")


#!/usr/bin/python
# -*- coding: utf-8 -*-
import os

from Qt import QtWidgets
from enum import Enum

from .pkg_info import __version__ as _v

__doc__ = """
| The :py:mod:`NodeGraphQt.constants` namespace contains variables and enums 
 used throughout the NodeGraphQt library.
"""

# ================================== PRIVATE ===================================

URI_SCHEME = 'nodegraphqt://'
URN_SCHEME = 'nodegraphqt::'

# === PATHS ===

BASE_PATH = os.path.dirname(os.path.abspath(__file__))
ICON_PATH = os.path.join(BASE_PATH, 'widgets', 'icons')
ICON_DOWN_ARROW = os.path.join(ICON_PATH, 'down_arrow.png')
ICON_NODE_BASE = os.path.join(ICON_PATH, 'node_base.png')

# === DRAW STACK ORDER ===

Z_VAL_PIPE = -1
Z_VAL_NODE = 1
Z_VAL_PORT = 2
Z_VAL_NODE_WIDGET = 3

# === ITEM CACHE MODE ===

# QGraphicsItem.NoCache
# QGraphicsItem.DeviceCoordinateCache
# QGraphicsItem.ItemCoordinateCache

ITEM_CACHE_MODE = QtWidgets.QGraphicsItem.DeviceCoordinateCache

# === NODE LAYOUT DIRECTION ===

#: Mode for vertical node layout.
NODE_LAYOUT_VERTICAL = 0
#: Mode for horizontal node layout.
NODE_LAYOUT_HORIZONTAL = 1
#: Variable for setting the node layout direction.
# NODE_LAYOUT_DIRECTION = NODE_LAYOUT_VERTICAL
NODE_LAYOUT_DIRECTION = NODE_LAYOUT_HORIZONTAL

# =================================== GLOBAL ===================================


class VersionEnum(Enum):
    """
    Current framework version.
    :py:mod:`NodeGraphQt.constants.VersionEnum`
    """
    #:
    VERSION = _v
    #:
    MAJOR = int(_v.split('.')[0])
    #:
    MINOR = int(_v.split('.')[1])
    #:
    PATCH = int(_v.split('.')[2])

# =================================== VIEWER ===================================


class ViewerEnum(Enum):
    """
    Node graph viewer styling layout:
    :py:mod:`NodeGraphQt.constants.ViewerEnum`
    """
    #: default background color for the node graph.
    # BACKGROUND_COLOR = (35, 35, 35)
    BACKGROUND_COLOR = (155, 155, 200)
    #: style node graph background with no grid or dots.
    GRID_DISPLAY_NONE = 0
    #: style node graph background with dots.
    GRID_DISPLAY_DOTS = 1
    #: style node graph background with grid lines.
    GRID_DISPLAY_LINES = 2
    #: grid size when styled with grid lines.
    GRID_SIZE = 50
    #: grid line color.
    GRID_COLOR = (45, 45, 45)


class ViewerNavEnum(Enum):
    """
    Node graph viewer navigation styling layout:
    :py:mod:`NodeGraphQt.constants.ViewerNavEnum`
    """
    #: default background color.
    BACKGROUND_COLOR = (25, 25, 25)
    #: default item color.
    ITEM_COLOR = (35, 35, 35)

# ==================================== NODE ====================================


class NodeEnum(Enum):
    """
    Node styling layout:
    :py:mod:`NodeGraphQt.constants.NodeEnum`
    """
    #: default node width.
    WIDTH = 160
    #: default node height.
    HEIGHT = 60
    #: default node icon size (WxH).
    ICON_SIZE = 18
    #: default node overlay color when selected.
    SELECTED_COLOR = (255, 255, 255, 30)
    #: default node border color when selected.
    # SELECTED_BORDER_COLOR = (254, 207, 42, 255)
    SELECTED_BORDER_COLOR = (52, 152, 219, 255 )

# ==================================== PORT ====================================


class PortEnum(Enum):
    """
    Port styling layout:
    :py:mod:`NodeGraphQt.constants.PortEnum`
    """
    #: default port size.
    SIZE = 22.0
    #: default port color. (r, g, b, a)
    COLOR = (49, 115, 100, 255)
    #: default port border color.
    BORDER_COLOR = (29, 202, 151, 255)
    #: port color when selected.
    ACTIVE_COLOR = (14, 45, 59, 255)
    #: port border color when selected.
    ACTIVE_BORDER_COLOR = (107, 166, 193, 255)
    #: port color on mouse over.
    HOVER_COLOR = (17, 43, 82, 255)
    #: port border color on mouse over.
    HOVER_BORDER_COLOR = (136, 255, 35, 255)
    #: threshold for selecting a port.
    CLICK_FALLOFF = 15.0


class PortTypeEnum(Enum):
    """
    Port connection types:
    :py:mod:`NodeGraphQt.constants.PortTypeEnum`
    """
    #: Connection type for input ports.
    IN = 'in'
    #: Connection type for output ports.
    OUT = 'out'

# ==================================== PIPE ====================================


class PipeEnum(Enum):
    """
    Pipe styling layout:
    :py:mod:`NodeGraphQt.constants.PipeEnum`
    """
    #: default width.
    WIDTH = 1.2
    ACTIVE_WIDTH = 2
    HIGHLIGHT_WIDTH = 2
    #: default color.
    COLOR = (175, 95, 30, 255)
    #: pipe color to a node when it's disabled.
    DISABLED_COLOR = (190, 20, 20, 255)
    #: pipe color when selected or mouse over.
    # ACTIVE_COLOR = (70, 255, 220, 255)
    ACTIVE_COLOR = (175, 95, 30, 255)
    #: pipe color to a node when it's selected.
    # HIGHLIGHT_COLOR = (232, 184, 13, 255)
    HIGHLIGHT_COLOR = (52, 152, 219 )
    #: draw connection as a line.
    DRAW_TYPE_DEFAULT = 0
    #: draw connection as dashed lines.
    DRAW_TYPE_DASHED = 1
    #: draw connection as a dotted line.
    DRAW_TYPE_DOTTED = 2


class PipeSlicerEnum(Enum):
    """
    Slicer Pipe styling layout:
    :py:mod:`NodeGraphQt.constants.PipeSlicerEnum`
    """
    #: default width.
    WIDTH = 1.5
    #: default color.
    COLOR = (255, 50, 75)


class PipeLayoutEnum(Enum):
    """
    Pipe connection drawing layout:
    :py:mod:`NodeGraphQt.constants.PipeLayoutEnum`
    """
    #: draw straight lines for pipe connections.
    STRAIGHT = 0
    #: draw curved lines for pipe connections.
    CURVED = 1
    #: draw angled lines for pipe connections.
    ANGLE = 2


# === PROPERTY BIN WIDGET ===

#: Property type will hidden in the properties bin (default).
NODE_PROP = 0
#: Property type represented with a QLabel widget in the properties bin.
NODE_PROP_QLABEL = 2
#: Property type represented with a QLineEdit widget in the properties bin.
NODE_PROP_QLINEEDIT = 3
#: Property type represented with a QTextEdit widget in the properties bin.
NODE_PROP_QTEXTEDIT = 4
#: Property type represented with a QComboBox widget in the properties bin.
NODE_PROP_QCOMBO = 5
#: Property type represented with a QCheckBox widget in the properties bin.
NODE_PROP_QCHECKBOX = 6
#: Property type represented with a QSpinBox widget in the properties bin.
NODE_PROP_QSPINBOX = 7
#: Property type represented with a ColorPicker widget in the properties bin.
NODE_PROP_COLORPICKER = 8
#: Property type represented with a Slider widget in the properties bin.
NODE_PROP_SLIDER = 9
#: Property type represented with a file selector widget in the properties bin.
NODE_PROP_FILE = 10
#: Property type represented with a file save widget in the properties bin.
NODE_PROP_FILE_SAVE = 11
#: Property type represented with a vector2 widget in the properties bin.
NODE_PROP_VECTOR2 = 12
#: Property type represented with vector3 widget in the properties bin.
NODE_PROP_VECTOR3 = 13
#: Property type represented with vector4 widget in the properties bin.
NODE_PROP_VECTOR4 = 14
#: Property type represented with float widget in the properties bin.
NODE_PROP_FLOAT = 15
#: Property type represented with int widget in the properties bin.
NODE_PROP_INT = 16
#: Property type represented with button widget in the properties bin.
NODE_PROP_BUTTON = 17



import os
from pathlib import Path
from typing import List, Tuple, Dict, Any

import cv2

from depthai_sdk.readers.abstract_reader import AbstractReader

_videoExt = ['.mjpeg', '.avi', '.mp4', '.h265', '.h264']


class VideoCapReader(AbstractReader):
    """
    Reads stream from mp4, mjpeg, h264, h265
    """
    initialFrames: Dict[str, Any] = dict()
    shapes: Dict[str, Tuple[int, int]] = dict()
    readers: Dict[str, cv2.VideoCapture] = dict()

    def __init__(self, path: Path) -> None:
        if path.is_file():
            stream = path.stem if (path.stem in ['left', 'right']) else 'color'
            self.readers[stream] = cv2.VideoCapture(str(path))
        else:
            for fileName in os.listdir(str(path)):
                f_name, ext = os.path.splitext(fileName)
                if ext not in _videoExt: continue
                stream = f_name if (f_name == 'left' or f_name == 'right') else 'color'
                self.readers[stream] = cv2.VideoCapture(str(path / fileName))

        for name, reader in self.readers.items():
            ok, f = reader.read()
            self.shapes[name] = (
                f.shape[1],
                f.shape[0]
            )
            self.initialFrames[name] = f

    def read(self):
        frames = dict()
        for name, reader in self.readers.items():
            if self.initialFrames[name] is not None:
                frames[name] = self.initialFrames[name].copy()
                self.initialFrames[name] = None

            if not self.readers[name].isOpened(): return False

            ok, frame = self.readers[name].read()
            if not ok: return False

            frames[name] = frame

        return frames

    def getStreams(self) -> List[str]:
        return [name for name in self.readers]

    def getShape(self, name: str) -> Tuple[int, int]:  # Doesn't work as expected!!
        return self.shapes[name]

    def close(self):
        [r.release() for _, r in self.readers.items()]

    def disableStream(self, name: str):
        if name in self.readers:
            self.readers.pop(name)



import copy
import time
import warnings
from pathlib import Path
from typing import Dict, Any, Optional, List, Union, Callable

import cv2
import depthai as dai

from depthai_sdk.args_parser import ArgsParser
from depthai_sdk.classes.output_config import BaseConfig, RecordConfig, OutputConfig, SyncConfig
from depthai_sdk.components.camera_component import CameraComponent
from depthai_sdk.components.component import Component
from depthai_sdk.components.imu_component import IMUComponent
from depthai_sdk.components.nn_component import NNComponent
from depthai_sdk.components.parser import parse_usb_speed
from depthai_sdk.components.stereo_component import StereoComponent
from depthai_sdk.oak_device import OakDevice
from depthai_sdk.record import RecordType, Record
from depthai_sdk.replay import Replay
from depthai_sdk.utils import configPipeline
from depthai_sdk.visualize import Visualizer


class UsbWarning(UserWarning):
    pass


class OakCamera:
    """
    OakCamera improves ease of use when developing apps for OAK devices.

    It abstracts DepthAI API pipeline building, different camera permutations, stream recording/replaying, it adds
    debugging features, does AI model handling, message syncing & visualization, and much more.

    It was designed with interoperability with depthai API in mind.
    """

    # User should be able to access these:
    _pipeline: dai.Pipeline = None
    _oak: OakDevice  # Init this object by default
    _args: Dict[str, Any] = None  # User defined arguments
    replay: Optional[Replay] = None

    _usb_speed: Optional[dai.UsbSpeed] = None
    _device_name: str = None  # MxId / IP / USB port

    # Whether to stop running the OAK camera. Used by oak.running()
    _stop: bool = False

    def __init__(self,
                 device: Optional[str] = None,  # MxId / IP / USB port
                 usbSpeed: Union[None, str, dai.UsbSpeed] = None,  # Auto by default
                 replay: Optional[str] = None,
                 rotation: int = 0,
                 args: Union[bool, Dict] = True
                 ):
        """
        Initializes OakCamera

        Args:
            device (str, optional): OAK device we want to connect to
            usb2 (bool, optional): Force USB2 mode
            replay (str, optional): Replay a depthai-recording - either local path, or from depthai-recordings repo
            args (None, bool, Dict): Use user defined arguments when constructing the pipeline
        """
        self._device_name = device
        self._usb_speed = parse_usb_speed(usbSpeed)
        self._oak = OakDevice()
        self._pipeline = dai.Pipeline()
        self._pipeline_built = False

        self._components: List[Component] = []  # List of components
        self._out_templates: List[BaseConfig] = []

        self._rotation = rotation

        if args:
            if isinstance(args, bool):
                if args:
                    self._args = ArgsParser.parseArgs()
                    # Set up the OakCamera
                    if self._args.get('recording', None):
                        replay = self._args.get('recording', None)
                    if self._args.get('deviceId', None):
                        self._device_name = self._args.get('deviceId', None)
                    if self._args.get('usbSpeed', None):
                        self._usb_speed = parse_usb_speed(self._args.get('usbSpeed', None))

                # else False - we don't want to parse user arguments
            else:  # Already parsed
                self._args = args

        if replay:
            self.replay = Replay(replay)
            print('Available streams from recording:', self.replay.getStreams())

    def create_camera(self,
                      source: str,
                      resolution: Union[
                          None, str, dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution] = None,
                      fps: Optional[float] = None,
                      encode: Union[None, str, bool, dai.VideoEncoderProperties.Profile] = None,
                      ) -> CameraComponent:
        """
        Creates Camera component. This abstracts ColorCamera/MonoCamera nodes and supports mocking the camera when
        recording is passed during OakCamera initialization. Mocking the camera will send frames from the host to the
        OAK device (via XLinkIn node).

        Args:
            source (str): Either 'color', 'left' or 'right' (these 2 will create MonoCamera nodes)
            resolution (str/SensorResolution): Sensor resolution of the camera.
            fps (float): Sensor FPS
            encode (bool/str/Profile): Whether we want to enable video encoding (accessible via cameraComponent.out_encoded). If True, it will use MJPEG
        """
        comp = CameraComponent(self._pipeline,
                               source=source,
                               resolution=resolution,
                               fps=fps,
                               encode=encode,
                               rotation=self._rotation,
                               replay=self.replay,
                               args=self._args, )
        self._components.append(comp)
        return comp

    def create_nn(self,
                  model: Union[str, Path],
                  input: Union[CameraComponent, NNComponent],
                  nn_type: Optional[str] = None,
                  tracker: bool = False,  # Enable object tracker - only for Object detection models
                  spatial: Union[None, bool, StereoComponent] = None,
                  decode_fn: Optional[Callable] = None,
                  ) -> NNComponent:
        """
        Creates Neural Network component.

        Args:
            model (str / Path): str for SDK supported model or Path to custom model's json/blob
            input (CameraComponent/NNComponent): Input to the model. If NNComponent (detector), it creates 2-stage NN
            nn_type (str): Type of the network (yolo/mobilenet) for on-device NN result decoding (only needed if blob path was specified)
            tracker: Enable object tracker, if model is object detector (yolo/mobilenet)
            spatial: Calculate 3D spatial coordinates, if model is object detector (yolo/mobilenet) and depth stream is available
            decode_fn: Custom decoding function for the model's output
        """
        comp = NNComponent(self._pipeline,
                           model=model,
                           input=input,
                           nn_type=nn_type,
                           tracker=tracker,
                           spatial=spatial,
                           decode_fn=decode_fn,
                           replay=self.replay,
                           args=self._args)
        self._components.append(comp)
        return comp

    def create_stereo(self,
                      resolution: Union[None, str, dai.MonoCameraProperties.SensorResolution] = None,
                      fps: Optional[float] = None,
                      left: Union[None, dai.Node.Output, CameraComponent] = None,  # Left mono camera
                      right: Union[None, dai.Node.Output, CameraComponent] = None,  # Right mono camera
                      ) -> StereoComponent:
        """
        Create Stereo camera component. If left/right cameras/component aren't specified they will get created internally.

        Args:
            resolution (str/SensorResolution): If monochrome cameras aren't already passed, create them and set specified resolution
            fps (float): If monochrome cameras aren't already passed, create them and set specified FPS
            left (CameraComponent/dai.node.MonoCamera): Pass the camera object (component/node) that will be used for stereo camera.
            right (CameraComponent/dai.node.MonoCamera): Pass the camera object (component/node) that will be used for stereo camera.
        """
        comp = StereoComponent(self._pipeline,
                               resolution=resolution,
                               fps=fps,
                               left=left,
                               right=right,
                               replay=self.replay,
                               args=self._args)
        self._components.append(comp)
        return comp

    def create_imu(self) -> IMUComponent:
        """
        Create IMU component
        """
        comp = IMUComponent(self._pipeline)
        self._components.append(comp)
        return comp

    def _init_device(self) -> None:
        """
        Connect to the OAK camera
        """
        if self._device_name:
            deviceInfo = dai.DeviceInfo(self._device_name)
        else:
            (found, deviceInfo) = dai.Device.getFirstAvailableDevice()
            if not found:
                raise Exception("No OAK device found to connect to!")

        version = self._pipeline.getOpenVINOVersion()

        if self._usb_speed == dai.UsbSpeed.SUPER:
            self._oak.device = dai.Device(
                version=version,
                deviceInfo=deviceInfo,
                usb2Mode=True
            )
        else:
            self._oak.device = dai.Device(
                version=version,
                deviceInfo=deviceInfo,
                maxUsbSpeed=dai.UsbSpeed.SUPER if self._usb_speed is None else self._usb_speed
            )

        # TODO test with usb3 (SUPER speed)
        if self._usb_speed != dai.UsbSpeed.HIGH and self._oak.device.getUsbSpeed() == dai.UsbSpeed.HIGH:
            warnings.warn("Device connected in USB2 mode! This might cause some issues. "
                          "In such case, please try using a (different) USB3 cable, "
                          "or force USB2 mode 'with OakCamera(usbSpeed=depthai.UsbSpeed.HIGH)'", UsbWarning)

    def config_camera(self, rotation: Optional[int] = None) -> None:
        """
        Configures general camera settings.
        Args:
            rotation: Rotate the camera output by this amount of degrees, 0 by default, 90, 180, 270 are supported.
        """
        self._rotation = rotation or self._rotation

    def config_pipeline(self,
                        xlinkChunk: Optional[int] = None,
                        calib: Optional[dai.CalibrationHandler] = None,
                        tuningBlob: Optional[str] = None,
                        openvinoVersion: Union[None, str, dai.OpenVINO.Version] = None
                        ):
        """
        Configures DepthAI pipeline.
        @param xlinkChunk: Chunk size of XLink messages. 0 can result in lower latency
        @param calib: Calibration data to be uploaded to OAK
        @param tuningBlob: Camera tuning blob
        @param openvinoVersion: Force specific OpenVINO version
        """
        configPipeline(self._pipeline, xlinkChunk, calib, tuningBlob, openvinoVersion)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        print("Closing OAK camera")
        if self.replay:
            print("Closing replay")
            self.replay.close()
        if self._oak.device is not None:
            self._oak.device.close()

        for out in self._out_templates:
            if isinstance(out, RecordConfig):
                out.rec.close()

    def start(self, blocking=False):
        """
        Start the application - upload the pipeline to the OAK device.
        Args:
            blocking (bool):  Continuously loop and call oak.poll() until program exits
        """
        if not self._pipeline_built:
            self.build()  # Build the pipeline

        self._oak.device.startPipeline(self._pipeline)

        self._oak.init_callbacks(self._pipeline)

        for xout in self._oak.oak_out_streams:  # Start FPS counters
            xout.start_fps()

        if self.replay:
            self.replay.createQueues(self._oak.device)
            # Called from Replay module on each new frame sent to the device.
            self.replay.start(self._oak.new_msg)

        # Check if callbacks (sync/non-sync are set)
        if blocking:
            # Constant loop: get messages, call callbacks
            while self.running():
                time.sleep(0.001)
                self.poll()

    def running(self) -> bool:
        """
        Check if camera is running.
        Returns:
            True if camera is running, False otherwise.
        """
        return not self._stop

    def poll(self):
        """
        Poll events; cv2.waitKey, send controls to OAK (if controls are enabled), update, check syncs.
        """
        key = cv2.waitKey(1)
        if key == ord('q'):
            self._stop = True
            return

        # TODO: check if components have controls enabled and check whether key == `control`

        self._oak.check_sync()

        if self.replay:
            if self.replay._stop:
                self._stop = True
                return

        if self.device.isClosed():
            self._stop = True

    def build(self) -> dai.Pipeline:
        """
        Connect to the device and build the pipeline based on previously provided configuration. Configure XLink queues,
        upload the pipeline to the device. This function must only be called once!  build() is also called by start().
        Return:
            Built dai.Pipeline
        """
        if self._pipeline_built:
            raise Exception('Pipeline can be built only once!')

        self._pipeline_built = True
        if self.replay:
            self.replay.initPipeline(self._pipeline)

        # First go through each component to check whether any is forcing an OpenVINO version
        # TODO: check each component's SHAVE usage
        for c in self._components:
            ov = c._forced_openvino_version()
            if ov:
                if self._pipeline.getRequiredOpenVINOVersion() and self._pipeline.getRequiredOpenVINOVersion() != ov:
                    raise Exception(
                        'Two components forced two different OpenVINO version!'
                        'Please make sure that all your models are compiled using the same OpenVINO version.'
                    )
                self._pipeline.setOpenVINOVersion(ov)

        if self._pipeline.getRequiredOpenVINOVersion() == None:
            # Force 2021.4 as it's better supported (blobconverter, compile tool) for now.
            self._pipeline.setOpenVINOVersion(dai.OpenVINO.VERSION_2021_4)

        # Connect to the OAK camera
        self._init_device()

        # Go through each component
        for component in self._components:
            # Update the component now that we can query device info
            component._update_device_info(self._pipeline, self._oak.device, self._pipeline.getOpenVINOVersion())

        # Create XLinkOuts based on visualizers/callbacks enabled

        # TODO: clean this up and potentially move elsewhere

        names = []
        for out in self._out_templates:
            xouts = out.setup(self._pipeline, self._oak.device, names)
            self._oak.oak_out_streams.extend(xouts)

        # User-defined arguments
        if self._args:
            self.config_pipeline(
                xlinkChunk=self._args.get('xlinkChunkSize', None),
                tuningBlob=self._args.get('cameraTuning', None),
                openvinoVersion=self._args.get('openvinoVersion', None),
            )

        return self._pipeline

    def sync(self, outputs: Union[Callable, List[Callable]], callback: Callable, visualize=False):
        """
        Synchronize multiple components outputs forward them to the callback.
        Args:
            outputs: Component output(s)
            callback: Where to send synced streams
            visualize: Whether to draw on the frames (like with visualize())
        """
        if isinstance(outputs, Callable):
            outputs = [outputs]  # to list

        self._out_templates.append(SyncConfig(outputs, callback))

    def record(self,
               outputs: Union[Callable, List[Callable]],
               path: str,
               record_type: RecordType = RecordType.VIDEO):
        """
        Record component outputs. This handles syncing multiple streams (eg. left, right, color, depth) and saving
        them to the computer in desired format (raw, mp4, mcap, bag..).
        Args:
            outputs (Component/Component output): Component output(s) to be recorded
            path: Folder path where to save these streams
            record_type: Record type
        """
        if isinstance(outputs, Callable):
            outputs = [outputs]  # to list

        for i in range(len(outputs)):
            if isinstance(outputs[i], Component):
                outputs[i] = outputs[i].out.main

        record = Record(Path(path).resolve(), record_type)
        self._out_templates.append(RecordConfig(outputs, record))
        return record

    def show_graph(self):
        """
        Shows DepthAI Pipeline graph, which can be useful when debugging. Builds the pipeline (oak.build()).
        """
        from depthai_sdk.components.pipeline_graph import PipelineGraph

        if not self._pipeline_built:
            self.build()  # Build the pipeline

        PipelineGraph(self._pipeline.serializeToJson()['pipeline'])

    def visualize(self,
                  output: Union[List, Callable, Component],
                  record_path: Optional[str] = None,
                  scale: float = None,
                  fps=False,
                  callback: Callable = None):
        """
        Visualize component output(s). This handles output streaming (OAK->host), message syncing, and visualizing.
        Args:
            output (Component/Component output): Component output(s) to be visualized. If component is passed, SDK will visualize its default output (out())
            record_path: Path where to store the recording (visualization window name gets appended to that path), supported formats: mp4, avi
            scale: Scale the output window by this factor
            fps: Whether to show FPS on the output window
            callback: Instead of showing the frame, pass the Packet to the callback function, where it can be displayed
        """
        if record_path and isinstance(output, List):
            raise ValueError('Recording visualizer is only supported for a single output.')

        visualizer = Visualizer(scale, fps)
        return self._callback(output, callback, visualizer, record_path)

    def _callback(self,
                  output: Union[List, Callable, Component],
                  callback: Callable,
                  visualizer: Visualizer = None,
                  record_path: Optional[str] = None):
        if isinstance(output, List):
            for element in output:
                self._callback(element, callback, visualizer, record_path)
            return

        if isinstance(output, Component):
            output = output.out.main

        visualizer = copy.deepcopy(visualizer) or Visualizer()
        self._out_templates.append(OutputConfig(output, callback, visualizer, record_path))
        return visualizer

    def callback(self, output: Union[List, Callable, Component], callback: Callable):
        """
        Create a callback for the component output(s). This handles output streaming (OAK->Host) and message syncing.
        Args:
            output: Component output(s) to be visualized. If component is passed, SDK will visualize its default output (out())
            callback: Handler function to which the Packet will be sent
        """
        self._callback(output, callback)

    def get_stats_report(self) -> Dict[str, Any]:
        """
        Get statistics for the pipeline.
        """
        if not self._pipeline_built:
            return {}

        return self._oak.stats_report()

    def get_info_report(self) -> Dict[str, Any]:
        """
        Get information about the device.
        """
        if not self._pipeline_built:
            return {}

        return self._oak.info_report()

    @property
    def device(self) -> dai.Device:
        """
        Returns dai.Device object. oak.built() has to be called before querying this property!
        """
        if not self._pipeline_built:
            raise Exception("OAK device wasn't booted yet, make sure to call oak.build() or oak.start()!")
        return self._oak.device



from typing import Optional, Union, Any, Dict

import cv2
import depthai as dai

from depthai_sdk.components.camera_component import CameraComponent
from depthai_sdk.components.component import Component
from depthai_sdk.components.parser import parse_cam_socket, parse_median_filter
from depthai_sdk.oak_outputs.xout import XoutDisparity, XoutDepth
from depthai_sdk.oak_outputs.xout_base import XoutBase, StreamXout
from depthai_sdk.replay import Replay
from depthai_sdk.visualize.configs import StereoColor


class StereoComponent(Component):
    # Users should have access to these nodes
    node: dai.node.StereoDepth

    left: Union[None, CameraComponent, dai.node.MonoCamera]
    right: Union[None, CameraComponent, dai.node.MonoCamera]

    @property
    def depth(self) -> dai.Node.Output:
        # Depth output from the StereoDepth node.
        return self.node.depth

    @property
    def disparity(self) -> dai.Node.Output:
        # Disparity output from the StereoDepth node.
        return self.node.disparity

    _replay: Optional[Replay]  # Replay module
    _resolution: Union[None, str, dai.MonoCameraProperties.SensorResolution]
    _fps: Optional[float]
    _args: Dict

    def __init__(self,
                 pipeline: dai.Pipeline,
                 resolution: Union[None, str, dai.MonoCameraProperties.SensorResolution] = None,
                 fps: Optional[float] = None,
                 left: Union[None, CameraComponent, dai.node.MonoCamera] = None,  # Left mono camera
                 right: Union[None, CameraComponent, dai.node.MonoCamera] = None,  # Right mono camera
                 replay: Optional[Replay] = None,
                 args: Any = None,
                 ):
        """
        Args:
            pipeline (dai.Pipeline): DepthAI pipeline
            resolution (str/SensorResolution): If monochrome cameras aren't already passed, create them and set specified resolution
            fps (float): If monochrome cameras aren't already passed, create them and set specified FPS
            left (None / dai.None.Output / CameraComponent): Left mono camera source. Will get handled by Camera object.
            right (None / dai.None.Output / CameraComponent): Right mono camera source. Will get handled by Camera object.
            replay (Replay object, optional): Replay
            args (Any, optional): Use user defined arguments when constructing the pipeline
        """
        super().__init__()
        self.out = self.Out(self)

        self._replay = replay
        self._resolution = resolution
        self._fps = fps
        self._args = args

        self.left = left
        self.right = right

        self.node = pipeline.createStereoDepth()
        self.node.setDefaultProfilePreset(dai.node.StereoDepth.PresetMode.HIGH_DENSITY)

        # Configuration variables
        self._colorize = StereoColor.GRAY
        self._colormap = cv2.COLORMAP_TURBO
        self._use_wls_filter = False
        self._wls_lambda = 8000
        self._wls_sigma = 1.5

    def _update_device_info(self, pipeline: dai.Pipeline, device: dai.Device, version: dai.OpenVINO.Version):
        if self._replay:
            if isinstance(self.left, CameraComponent):
                self.left = self.left.node  # CameraComponent -> node
            if isinstance(self.right, CameraComponent):
                self.right = self.right.node  # CameraComponent -> node

            self._replay.initStereoDepth(self.node)
        else:
            # TODO: check sensor names / device name whether it has stereo camera pair (or maybe calibration?)
            if len(device.getCameraSensorNames()) == 1:
                raise Exception('OAK-1 camera does not have Stereo camera pair!')

            if not self.left:
                self.left = CameraComponent(pipeline, 'left', self._resolution, self._fps, replay=self._replay)
                self.left._update_device_info(pipeline, device, version)
            if not self.right:
                self.right = CameraComponent(pipeline, 'right', self._resolution, self._fps, replay=self._replay)
                self.right._update_device_info(pipeline, device, version)

            if isinstance(self.left, CameraComponent):
                self.left = self.left.node  # CameraComponent -> node
            if isinstance(self.right, CameraComponent):
                self.right = self.right.node  # CameraComponent -> node

            # TODO: use self._args to setup the StereoDepth node
            # Connect Mono cameras to the StereoDepth node
            self.left.out.link(self.node.left)
            self.right.out.link(self.node.right)

            if 0 < len(device.getIrDrivers()):
                laser = self._args.get('irDotBrightness', None)
                laser = int(laser) if laser else 800
                if 0 < laser:
                    device.setIrLaserDotProjectorBrightness(laser)
                    print(f'Setting IR laser dot projector brightness to {laser}mA')

                led = self._args.get('irFloodBrightness', None)
                if led is not None:
                    device.setIrFloodLightBrightness(int(led))
                    print(f'Setting IR flood LED brightness to {int(led)}mA')

        if self._args:
            self._config_stereo_args(self._args)

    def _config_stereo_args(self, args: Dict):
        if not isinstance(args, Dict):
            args = vars(args)  # Namespace -> Dict
        self.config_stereo(
            confidence=args.get('disparityConfidenceThreshold', None),
            median=args.get('stereoMedianSize', None),
            extended=args.get('extendedDisparity', None),
            subpixel=args.get('subpixel', None),
            lrCheck=args.get('lrCheck', None),
            sigma=args.get('sigma', None),
            lrCheckThreshold=args.get('lrcThreshold', None),
        )

    def config_stereo(self,
                      confidence: Optional[int] = None,
                      align: Union[None, str, dai.CameraBoardSocket] = None,
                      median: Union[None, int, dai.MedianFilter] = None,
                      extended: Optional[bool] = None,
                      subpixel: Optional[bool] = None,
                      lrCheck: Optional[bool] = None,
                      sigma: Optional[int] = None,
                      lrCheckThreshold: Optional[int] = None,
                      ) -> None:
        """
        Configures StereoDepth modes and options.
        """
        if confidence: self.node.initialConfig.setConfidenceThreshold(confidence)
        if align: self.node.setDepthAlign(parse_cam_socket(align))
        if median: self.node.setMedianFilter(parse_median_filter(median))
        if extended: self.node.initialConfig.setExtendedDisparity(extended)
        if subpixel: self.node.initialConfig.setSubpixel(subpixel)
        if lrCheck: self.node.initialConfig.setLeftRightCheck(lrCheck)
        if sigma: self.node.initialConfig.setBilateralFilterSigma(sigma)
        if lrCheckThreshold: self.node.initialConfig.setLeftRightCheckThreshold(lrCheckThreshold)

    def configure_postprocessing(self,
                                 colorize: StereoColor = None,
                                 colormap: int = None,
                                 wls_filter: bool = None,
                                 wls_lambda: float = None,
                                 wls_sigma: float = None) -> None:
        self._colorize = colorize or self._colorize
        self._colormap = colormap or self._colormap
        self._use_wls_filter = wls_filter or self._use_wls_filter
        self._wls_lambda = wls_lambda or self._wls_lambda
        self._wls_sigma = wls_sigma or self._wls_sigma

    def _get_disparity_factor(self, device: dai.Device) -> float:
        """
        Calculates the disparity factor used to calculate depth from disparity.
        `depth = disparity_factor / disparity`
        @param device: OAK device
        """
        calib = device.readCalibration()
        baseline = calib.getBaselineDistance(useSpecTranslation=True) * 10  # mm
        intrinsics = calib.getCameraIntrinsics(dai.CameraBoardSocket.RIGHT, self.right.getResolutionSize())
        focalLength = intrinsics[0][0]
        disp_levels = self.node.getMaxDisparity() / 95
        return baseline * focalLength * disp_levels

    """
    Available outputs (to the host) of this component
    """

    class Out:
        _comp: 'StereoComponent'

        def __init__(self, stereoComponent: 'StereoComponent'):
            self._comp = stereoComponent

        def main(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutBase:
            # By default, we want to show disparity
            return self.depth(pipeline, device)

        def disparity(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutBase:
            fps = self._comp.left.getFps() if self._comp._replay is None else self._comp._replay.getFps()

            out = XoutDisparity(
                disparity_frames=StreamXout(self._comp.node.id, self._comp.disparity),
                mono_frames=StreamXout(self._comp.node.id, self._comp.right.out),
                max_disp=self._comp.node.getMaxDisparity(),
                fps=fps,
                colorize=self._comp._colorize,
                colormap=self._comp._colormap,
                use_wls_filter=self._comp._use_wls_filter,
                wls_lambda=self._comp._wls_lambda,
                wls_sigma=self._comp._wls_sigma
            )

            return self._comp._create_xout(pipeline, out)

        def depth(self, pipeline: dai.Pipeline, device: dai.Device) -> XoutBase:
            fps = self._comp.left.getFps() if self._comp._replay is None else self._comp._replay.getFps()
            out = XoutDepth(
                device=device,
                frames=StreamXout(self._comp.node.id, self._comp.depth),
                mono_frames=StreamXout(self._comp.node.id, self._comp.right.out),
                fps=fps,
                colorize=self._comp._colorize,
                colormap=self._comp._colormap,
                use_wls_filter=self._comp._use_wls_filter,
                wls_lambda=self._comp._wls_lambda,
                wls_sigma=self._comp._wls_sigma
            )
            return self._comp._create_xout(pipeline, out)

    out: Out



import depthai as dai
from typing import Optional
from depthai_sdk.oak_outputs.xout_base import XoutBase, ReplayStream
from abc import ABC, abstractmethod


class Component(ABC):
    """
    SDK component is used as an abstraction to the current DepthAI API node or group of nodes.    
    """

    def _forced_openvino_version(self) -> Optional[dai.OpenVINO.Version]:
        """
        Checks whether the component forces a specific OpenVINO version. Only used by NNComponent (which overrides this
        method). This function is called after Camera has been configured and right before we connect to the OAK camera.
        @return: Forced OpenVINO version (optional).
        """
        return None

    @abstractmethod
    def _update_device_info(self, pipeline: dai.Pipeline, device: dai.Device, version: dai.OpenVINO.Version):
        """
        This function will be called after the app connects to the Device
        """
        raise NotImplementedError("Every component needs to include 'updateDeviceInfo()' method!")

    def _stream_name_ok(self, pipeline: dai.Pipeline, name: str) -> bool:
        # Check if there's already an XLinkOut stream with this name
        for node in pipeline.getAllNodes():
            if isinstance(node, dai.node.XLinkOut) and node.getStreamName() == name:
                return False
        return True

    def _create_xout(self, pipeline: dai.Pipeline, xout: XoutBase) -> XoutBase:
        for xstream in xout.xstreams():
            if not self._stream_name_ok(pipeline, xstream.name):
                continue

            if isinstance(xstream, ReplayStream):
                continue

            xlink = pipeline.createXLinkOut()
            xlink.setStreamName(xstream.name)
            xstream.stream.link(xlink.input)

        return xout



"""
General (standarized) NN outputs, to be used for higher-level abstractions (eg. automatic visualization of results).
"SDK supported NN models" will have to have standard NN output, so either dai.ImgDetections, or one of the outputs
below. If the latter, model json config will incldue handler.py logic for decoding to the standard NN output.
These will be integrated into depthai-core, bonus points for on-device decoding of some popular models.
"""
from dataclasses import dataclass
from typing import List, Tuple, Any

import numpy as np
from depthai import NNData, ImgDetection


class GenericNNOutput:
    """
    Generic NN output, to be used for higher-level abstractions (eg. automatic visualization of results).
    """

    def __init__(self, nn_data: NNData):
        self.nn_data = nn_data

# First we have Object detection results, which are already standarized with dai.ImgDetections

@dataclass
class Detections(GenericNNOutput):
    """
    Detection results containing bounding boxes, labels and confidences. Optionally can contain rotation angles.
    """

    def __init__(self, nn_data: NNData, is_rotated: bool = False):
        GenericNNOutput.__init__(self, nn_data)

        self.detections = []
        self.is_rotated = is_rotated
        if is_rotated:
            self.angles = []

    def add(self, label: int, confidence: float, bbox: Tuple[float, ...], angle: int = 0) -> None:
        det = ImgDetection()
        det.label = label
        det.confidence = confidence
        det.xmin = bbox[0]
        det.ymin = bbox[1]
        det.xmax = bbox[2]
        det.ymax = bbox[3]
        self.detections.append(det)
        if self.is_rotated:
            self.angles.append(angle)


@dataclass
class SemanticSegmentation(GenericNNOutput):  # In core, extend from NNData
    """
    Semantic segmentation results, with a mask for each class.

    Examples: `DeeplabV3`, `Lanenet`, `road-semgentation-adas-0001`.
    """
    mask: List[np.ndarray]  # 2D np.array for each class

    def __init__(self, nn_data: NNData, mask: List[np.ndarray]):
        super().__init__(nn_data)
        self.mask: List[np.ndarray] = mask


@dataclass
class ImgLandmarks(GenericNNOutput):  # In core, extend from NNData
    """
    Landmarks results, with a list of landmarks and pairs of landmarks to draw lines between.

    Examples: `human-pose-estimation-0001`, `openpose2`, `facial-landmarks-68`, `landmarks-regression-retail-0009`.
    """
    landmarks: List[List[Any]]
    pairs: List[Tuple[int, int]]  # Pairs of landmarks, to draw lines between them
    colors: List[Tuple[int, int, int]]  # Color for each landmark (eg. both elbows are in the same color)

    def __init__(self,
                 nn_data: NNData,
                 landmarks: List[List[Any]] = None,
                 pairs: List[Tuple[int, int]] = None,
                 colors: List[Tuple[int, int, int]] = None):
        super().__init__(nn_data)
        self.landmarks = landmarks
        self.pairs = pairs
        self.colors = colors


@dataclass
class InstanceSegmentation(GenericNNOutput):
    """
    Instance segmentation results, with a mask for each instance.
    """
    masks: List[np.ndarray]  # 2D np.array for each instance
    labels: List[int]  # Class label for each instance

    def __init__(self, nn_data: NNData, masks: List[np.ndarray], labels: List[int]):
        super().__init__(nn_data)

